
################################################################################
# IIPyhonUtils
#
# A library of Python utility classes developped by the
# Interactive Information Group of the Institute for Information Technology
# at the National Research Council of Canada (NRC).
#
#
# (c) National Research Council of Canada, 2000
# taken for unimacro/natlink tests by Quintijn Hoogenboom march
# adapt for Python2/Python3 integration. Quintijn april 2018
################################################################################
import os
import re
import types
import unittest
from pprint import pformat
import filecmp
import itertools
from natlinkcore import readwritefile

class AssertStringEqualsFailed(AssertionError):
    def __init__(self, exp_string, got_string, first_diff, orig_message):
        self.exp_string = exp_string
        self.got_string = got_string
        self.first_diff = first_diff
        self.orig_message = orig_message

    def __str__(self):
        if (self.first_diff < 0):
            format = "%s\nOne string was empty but not the other.\n" + \
                     "   Expected: \n%s\n" + \
                     "   Got: \n%s\n"
            message = format % (self.orig_message, repr(self.exp_string), repr(self.got_string))
        else:
            exp_with_tilde = self.exp_string[:self.first_diff] + '~~~' + self.exp_string[self.first_diff:]
            got_with_tilde = self.got_string[:self.first_diff] + '~~~' + self.got_string[self.first_diff:]
            format = "%s\n----\nStrings differed. See triple tilde (~~~) position below.\n" + \
                      "Expected:\n" + \
                      "   '%s'\n" + \
                      "Got:\n" + \
                      "   '%s'\n"
            message = format % (self.orig_message, repr(exp_with_tilde), repr(got_with_tilde))
        return message

class TestCaseWithHelpers(unittest.TestCase):
    """A subclass of pyUnit TestCase, that has helper method to
    facilitate unit testing.
    """

    def __init__(self, name):
        unittest.TestCase.__init__(self, name)

    def remind_me_to_implement_this_test(self):
        self.fail("DON'T FORGET TO IMPLEMENT THIS TEST!!!")

    def assert_equal(self, expected, got, mess="", epsilon=0):

        try:
            self.assert_equal_dispatch_according_to_type(expected, got, mess, epsilon)
        except RuntimeError as err:
            err = self.check_for_infinite_recursion_error(err)
            raise err

    def assert_equal_dispatch_according_to_type(self, expected, got, mess, epsilon):
        self.assertTrue(self.isnumber(epsilon), "assert_equal called with a value of epsilon that was not a number. Type of epsilon was: %s, value was %s" % (self.what_class(epsilon), epsilon))
        class_exp, class_got = self.what_class(expected), self.what_class(got)
        if class_exp != class_got:
            if self.isnumber(expected) and self.isnumber(got):
                pass
            elif self.isstringlike(expected) and self.isstringlike(got):
                pass
                # print('got PY2 str, expected unicode: %s'% got)
            else:
                mess = mess + "\n----\nThe two objects were not of same type"
                mess = mess + "\nExpected type: %s, Got type:  %s" % (type(expected), type(got))
                mess = mess + "\nExpected:\n   %s\nGot:\n   %s" % (expected, got)

                self.fail(mess)
        if type(expected) is None and type(got) is None:
            return

        if (self.issequence(expected)):
            self.assert_equal_sequence(expected, got, mess, epsilon)
        elif (self.isset(expected)):
            self.assert_equal_set(expected, got, mess, epsilon)
        elif (self.isdictionary(expected)):
            self.assert_equal_dictionary(expected, got, mess, epsilon)
        elif (self.isnumber(expected)):
            mess = mess + "\n----\nThe two numbers differed significantly."
            mess = mess + "\nExpected:\n   %s\nGot:\n   %s" % (expected, got)
            if epsilon == 0:
                self.assertTrue(expected == got, mess)
            else:
                self.assertTrue((got <= expected + epsilon) and (got >= expected - epsilon),
                         mess + "\nValue was not within epsilon=%s range of expected value." % epsilon)
        elif (self.isstring(expected)):
            self.assert_equal_string(expected, got, mess)
        elif (self.isbasetype(expected)):
            self.assertTrue(expected == got, mess)
        else:
            self.assert_equal_objects(expected, got, mess, epsilon)


    def assert_equal_string(self, exp_string, got_string, message=''):
        if self.isbinary(got_string):
            mess = "WARNING, got binary string expectied str, %s "% got_string
            got_string = str(got_string)
        first_diff = self.find_first_diff_char(got_string, exp_string)

        if first_diff != None:
            raise AssertStringEqualsFailed(exp_string, got_string, first_diff, message)

    def assert_equal_sequence(self, expected, got, mess="", epsilon=0):
        mess = mess + "\n----\nThe two sequences differred\nExpected sequence:\n   %s\nGot sequence:\n   %s" % \
               (pformat(expected), pformat(got))

        if len(expected) != len(got):
            mess = mess + "\nlengths differ, len(expected): %s, len(got): %s"% (len(expected), len(got))
            if len(expected) > len(got):
                mess = mess + "\nextra in expected: %s"% expected[len(got):]
            else:
                mess = mess + "\nextra in got: %s"% got[len(expected):]
        self.assert_equal(len(expected), len(got), mess + "\nThe two sequences did not have the same length.")


        for ii in range(len(expected)):
            mess_ii = mess + "\nThe two sequences differed at element ii=%s." % (ii)
            self.assert_equal(expected[ii], got[ii], mess_ii, epsilon)

    def assert_equal_set(self, expected, got, mess="", epsilon=0):
        mess = mess + "\n----\nThe two sets differ\nExpected sequence:\n   %s\nGot sequence:\n   %s" % \
               (pformat(expected), pformat(got))
        union = expected & got
        restexpected = expected - union
        restgot = got - union

        if union != expected:
            mess = mess + "\nonly in expected: %s"% restexpected
        if union != got:
            mess = mess + "\nonly in got: %s"% restgot
        self.assertEqual(expected, got, mess)


    def assert_equal_dictionary(self, expected, got, mess="", epsilon=0):
        expected_keys = set(expected.keys())
        got_keys = set(got.keys())
        # expected_keys = sorted(expected.keys())
        # got_keys = sorted(got.keys())

        mess = mess + "\n----\nThe two dictionaries differed"
        mess = mess + "\nExpected:\n   %s\nGot:\n   %s" % (pformat(expected), pformat(got))

        # NOTE: Keys must be exactly equal, even if epsilon is greater than 0.
        self.assert_equal(expected_keys, got_keys,
                              mess + "\nDictionaries did not have the same list of keys.")
        for a_key in expected_keys:
            self.assert_equal(expected[a_key], got[a_key],
                    mess + "\nValues of dictionaries differed at key '%s'" % a_key,
                    epsilon)


    def assert_equal_objects(self, expected, got, mess, epsilon=0):
        self.assert_equal(self.what_class(expected), self.what_class(got),
                           mess + "\n----\nThe two objects were not of the same class or type.")
        
        result = expected == got
        self.assert_equal(True, result,
                           mess + "\n----\nObjects differ"
                                + "\nExpected:\n   %s\nGot:\n   %s" % (repr(expected), repr(got)))

    def check_for_infinite_recursion_error(self, err):
        if (isinstance(err, RuntimeError) and
            err.args == ('maximum recursion depth exceeded', )):
            err.args = \
               ("maximum recursion depth exceeded.\n" + \
                "Error happened while doing an assert_equal().\n" + \
                "Maybe one of the arguments of assert_equal() has an infinite loop in its composition structure?",
                )
        return err

    def assert_not_equal(self, not_expected, got, mess):
        if got == not_expected:
            mess = mess + "\nValues were equal when they should not have been. Got: %s" % got
            self.fail(mess)

    def assert_string_contains(self, pattern, the_string, mess=''):
        self.assertTrue(the_string.find(pattern) != -1,
                     mess + "\nSubstring: '%s' was not found in string: '%s'" % (pattern, the_string))

    def assert_sequences_have_same_length(self, expected, got, mess):
        display_both_lists_mess = \
           "\nExpected list:\n   %s\nGot list:\n   %s" % (pformat(expected), pformat(got))

        if len(expected) != len(got):
            self.fail(mess + "\nExpeted sequence of length %s, but got sequence of length %s"
                      % (len(expected), len(got)) + \
                      display_both_lists_mess)

    def assert_equal_files(self, expected_file, actual_file, mess):
        """check contents of both files, to be equal

        a small issue (class=copyright) is in here because of the WhatCanISayTest files...

        """
        trunk, ext = os.path.splitext(str(expected_file))
        if ext in [".html", ".htm", ".txt", ".js", ".css", ".ini", ".log"]:
            encodingexp, bomexp, contentexp = readwritefile.readAnything(expected_file)
            encodinggot, bomgot, contentgot = readwritefile.readAnything(actual_file)
            if encodinggot != encodingexp:
                print("warning: different encoding found\n---exp: %s: %s\n+++got: %s: %s"%
                                  (expected_file, encodingexp, actual_file, encodinggot))
            if bomgot != bomexp:
                print("warning: different bom found\n---exp: %s: %s\n+++got: %s: %s"%
                                  (expected_file, bomexp, actual_file, bomgot))
            # 
            # 
            # with open(expected_file) as k:
            #     with open(actual_file) as l:
            gotlines = contentgot.split('\n')
            explines = contentexp.split('\n')
            for i, (kline, lline) in enumerate(zip(gotlines, explines)):
                if ext in [".html", ".htm"] and \
                        kline.find('class="copyright"') >= 0 and \
                        lline.find('class="copyright"') >= 0:
                    continue
                self.assert_equal(kline, lline, mess + '\nthe two files------------\n%s\nand\n %s should have been equal\nThey differ in line %s'%
                            (expected_file, actual_file, i+1))
        else:
            if not filecmp.cmp(expected_file, actual_file):
                message = mess + '\nthe two files------------\n%s\nand\n%s should have been equal\n--- comparison with filecmp.cmp ---'% \
                                    (expected_file, actual_file)
            self.fail(message)

    def assert_dicts_have_same_keys(self, expected_dict, got_dict, mess):
        expected_keys = list(expected_dict.keys())
        expected_keys.sort()
        got_keys = list(got_dict.keys())
        got_keys.sort()
        self.assert_equal(expected_keys, got_keys,
                                               "%s\nThe two dictionaries did not have the same keys" % mess)

    def find_first_diff_char(self, string1, string2):

        if string1 == None:
            string1 = ''
        if string2 == None:
            string2 = ''

        if (string1 == string2):
            return None
        if (len(string1) < len(string2)):
            upto = len(string1)-1;
        else:
            upto = len(string2)-1;
        if (upto < 0):
            # one of the strings was empty but not the other
            return -1

        for ii in range(upto+1):
            char1 = string1[ii]
            char2 = string2[ii]
            if char1 != char2:
                return ii;
        if (upto < len(string1)-1 or upto < len(string2)-1):
            return upto + 1;
        else:
            return undef;



    def assert_file_exists(self, fpath, mess):
        if not os.path.exists(fpath):
            self.fail(mess + "\nFile '%s' did not exist." % fpath)
        elif not os.path.isfile(fpath):
            self.fail(mess + "\nPath '%s' is not a file." % fpath)

    def assert_file_content_is(self, fpath, expected_content, message):
        self.assert_file_exists(fpath, message)
        file = open(fpath, 'r')
        got_content = file.read()
        file.close()
        self.assert_equal(expected_content, got_content, message)

    def assert_file_contains_N_lines(self, fpath, exp_N_lines, mess):
        self.assert_file_exists(fpath, mess)
        file = open(fpath, 'r')
        content = file.read()
        file.close()
        got_N_lines = len(content.split("\n"))
        self.assert_equal(exp_N_lines, got_N_lines, mess + "\nNumber of lines in the file was not as expected.")

    def make_empty_dir_for_file(self, fpath):
        directory = self.makedirs_for_file(fpath)
        files = os.listdir(directory)
        for file in files:
            if os.path.isfile(file):
                os.remove(file)
            elif os.path.isdir(file):
                os.removedirs(file)
            else:
                pass

    def makedirs_for_file(self, fpath):
        head, tail = os.path.split(fpath)
        try:
            os.makedirs(head)
        except:
            # Presumably, directory already existed
            pass
        return head

######################################################################
# Some introspection methods needed by TestCaseWithHelpers.
# We put them here instead of in a separate class, in order to make
# TestCaseWithHelpers self-contained.
######################################################################

    def what_class(self, instance):
        """Returns a string describing the class of an instance.

        It works with any Python class or Python standard data types (int, float,
        string, etc.), but not with extension classes."""

        is_class = 'unknown'
        try:
            tmp = instance.__class__
            is_class = tmp
        except AttributeError:
            #
            # The instance is not a python class. Maybe one of the
            # standard python data types?
            #
            is_class = type(instance)

        return str(is_class)

    def isnumber(self, instance):
        return isinstance(instance, int) or isinstance(instance, float)

    def issequence(self, instance):
        return isinstance(instance, list) or \
               isinstance(instance, tuple)

    def isset(self, instance):
        return isinstance(instance, set)

    def isdictionary(self, instance):
        return isinstance(instance, dict)

    # new QH, conversion python 2 to python 3:
    def isstring(self, instance):
        return type(instance) == str
    def isbinary(self, instance):
        return type(instance) == bytes

    def isstringlike(self, instance):
        return type(instance) == str


    def isbasetype(self, instance):
        return re.search("^\<type ", repr(self.what_class(instance)))

######################################################################
# Old deprecated names for some methods. Still supported for backward
# compatibility.
######################################################################

    def assert_equals(self, expected, got, mess, epsilon=0):
        print("\nWARNING: Call to deprecated method TestCaseWithHelpers.assert_equals()\n")
        self.assert_equal(expected, got, mess, epsilon)

    def assert_string_equals(self, exp_string, got_string, message=''):
        print("\nWARNING: Call to deprecated method TestCaseWithHelpers.assert_string_equals\n")
        self.assert_equal_string(exp_string, got_string, message)

    def assert_dicts_have_same_content(self, expected_dict, got_dict, mess):
        print("\nWARNING: Call to deprecated method TestCaseWithHelpers.assert_dicts_have_same_content\n")
        self.assert_equals(expected_dict, got_dict, mess)

    def assert_sequences_have_same_content(self, expected, got, mess):
        print("\nWARNING: Call to deprecated method TestCaseWithHelpers.assert_sequences_have_same_content\n")
        self.assert_equal(expected, got, mess)

#------------------------------------------------------------
def assert_equal(expected, got, mess="", epsilon=0):
    if expected == got: return
    try:
        assert_equal_dispatch_according_to_type(expected, got, mess, epsilon)
    except RuntimeError as err:
        err = self.check_for_infinite_recursion_error(err)
        raise err

def assert_equal_dispatch_according_to_type(expected, got, mess, epsilon):
    assert isnumber(epsilon), "assert_equal called with a value of epsilon that was not a number. Type of epsilon was: %s, value was %s" % (what_class(epsilon), epsilon)
    class_exp, class_got = what_class(expected), what_class(got)
    if class_exp != class_got:
        if self.isnumber(expected) and self.isnumber(got):
            pass
        elif self.isstringlike(expected) and self.isstringlike(got):
            pass
            # print('got PY2 str, expected unicode: %s'% got)
        else:
            mess = mess + "\n----\nThe two objects were not of same type"
            mess = mess + "\nExpected type: %s, Got type:  %s" % (type(expected), type(got))
            mess = mess + "\nExpected:\n   %s\nGot:\n   %s" % (expected, got)

            raise ValueError(mess)

    if (issequence(expected)):
        assert_equal_sequence(expected, got, mess, epsilon)
    elif (isset(expected)):
        assert_equal_set(expected, got, mess, epsilon)
    elif (isdictionary(expected)):
        assert_equal_dictionary(expected, got, mess, epsilon)
    elif (isnumber(expected)):
        mess = mess + "\n----\nThe two numbers differed significantly."
        mess = mess + "\nExpected:\n   %s\nGot:\n   %s" % (expected, got)
        if epsilon == 0:
            assert expected == got, mess
        else:
            assert (got <= expected + epsilon) and (got >= expected - epsilon), mess + "\nValue was not within epsilon=%s range of expected value." % epsilon
    elif (isstring(expected)):
        assert_equal_string(expected, got, mess)
    elif (isbasetype(expected)):
        assert expected == got, mess
    else:
        assert_equal_objects(expected, got, mess, epsilon)


def assert_equal_string(exp_string, got_string, mess=''):
    if isbinary(exp_string):
        raise ValueError("expected string should be unicode: %s"% exp_string)
    
    if isbinary(got_string):
        mess = "WARNING, got binary string expectied str, %s "% got_string
        got_string = utilsqh.convertToUnicode(got_string)
    if exp_string == got_string: return
    
    first_diff = find_first_diff_char(got_string, exp_string)

    if first_diff != None:
        expch = exp_string[:first_diff] + "~~~" + exp_string[first_diff:]
        gotch = got_string[:first_diff] + "~~~" + got_string[first_diff:]
        mess = mess + "\nStrings differ at position %s"% first_diff
        mess = mess + "\nexp: |%s|"% expch
        mess = mess + "\ngot: |%s|"% gotch
        raise ValueError(mess)

def assert_equal_sequence(expected, got, mess="", epsilon=0):
    if expected == got: return
    mess = mess + "\n----\nThe two sequences differred\nExpected sequence:\n   %s\nGot sequence:\n   %s" % \
           (pformat(expected), pformat(got))

    if len(expected) != len(got):
        mess = mess + "\nlengths differ, len(expected): %s, len(got): %s"% (len(expected), len(got))
        if len(expected) > len(got):
            mess = mess + "\nextra in expected: %s"% expected[len(got):]
        else:
            mess = mess + "\nextra in got: %s"% got[len(expected):]
    for i in range(min(len(expected), len(got))):
        if expected[i] != got[i]:
            mess = mess + "\nsequences differ at index %s"% i
            mess = mess + "\nexpected[%s]: |%s|"% (i, expected[i])
            mess = mess + "\n     got[%s]: [%s]"% (i, got[i])
            break
    raise ValueError(mess)

def assert_equal_set(expected, got, mess="", epsilon=0):
    if expected == got:
        return
    mess = mess + "\n----\nThe two sets differ\nExpected sequence:\n   %s\nGot sequence:\n   %s" % \
           (pformat(expected), pformat(got))
    union = expected & got
    restexpected = expected - union
    restgot = got - union

    if union != expected:
        mess = mess + "\nonly in expected: %s"% restexpected
    if union != got:
        mess = mess + "\nonly in got: %s"% restgot
    raise ValueError(mess)

def assert_equal_dictionary(expected, got, mess="", epsilon=0):
    """stand alone version of test
    """
    if expected == got: return
    expected_keys = set(expected.keys())
    got_keys = set(got.keys())

    mess = mess + "\n----\nThe two dictionaries differ"
    mess = mess + "\nExpected:\n   %s\nGot:\n   %s" % (pformat(expected), pformat(got))

    # NOTE: Keys must be exactly equal, even if epsilon is greater than 0.
    assert_equal_set(expected_keys, got_keys,
                          mess + "\nDictionaries did not have the same set of keys.")
    for a_key in expected_keys:
        assert_equal(expected[a_key], got[a_key],
                mess + "\nValues of dictionaries differed at key '%s'" % a_key,
                epsilon)

def find_first_diff_char(string1, string2):

    if string1 == None:
        string1 = ''
    if string2 == None:
        string2 = ''

    if (string1 == string2):
        return None
    if (len(string1) < len(string2)):
        upto = len(string1)-1;
    else:
        upto = len(string2)-1;
    if (upto < 0):
        # one of the strings was empty but not the other
        return -1

    for ii in range(upto+1):
        char1 = string1[ii]
        char2 = string2[ii]
        if char1 != char2:
            return ii;
    if (upto < len(string1)-1 or upto < len(string2)-1):
        return upto + 1;
    else:
        return undef;

def isnumber(instance):
    return isinstance(instance, int) or isinstance(instance, float)

def issequence(instance):
    return isinstance(instance, list) or \
           isinstance(instance, tuple)

def isset(instance):
    return isinstance(instance, set)

def isdictionary(instance):
    return isinstance(instance, dict)

# new QH, conversion python 2 to python 3:
def isstring(instance):
    return type(instance) == str
def isbinary(instance):
    return type(instance) == bytes

def isstringlike(instance):
    return type(instance) == str

def what_class(instance):
    """Returns a string describing the class of an instance.

    It works with any Python class or Python standard data types (int, float,
    string, etc.), but not with extension classes."""

    is_class = 'unknown'
    try:
        tmp = instance.__class__
        is_class = tmp
    except AttributeError:
        #
        # The instance is not a python class. Maybe one of the
        # standard python data types?
        #
        is_class = type(instance)

    return is_class

if __name__ == "__main__":
    import sys
    print("version", sys.version)
    print("test expects to go wrong!!!!!")
    # print(assert_equal("hello", "helloo"))
    # assert_equal([1,2,3], [1,2,4])
    d1 = dict(a=1, b=2, c=3)
    d2 = dict(a=1, b=2, c=4)
    assert_equal(d1, d2)


# (unimacro - natlink macro wrapper/extensions)
# (c) copyright 2003 Quintijn Hoogenboom (quintijn@users.sourceforge.net)
#                    Ben Staniford (ben_staniford@users.sourceforge.net)
#                    Bart Jan van Os (bjvo@users.sourceforge.net)
#
# This file is part of a SourceForge project called "unimacro" see
# http://unimacro.SourceForge.net).
#
# "unimacro" is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, see:
# http://www.gnu.org/licenses/gpl.txt
#
# "unimacro" is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; See the GNU General Public License details.
#
# "unimacro" makes use of another SourceForge project "natlink",
# which has the following copyright notice:
#
# Python Macro Language for Dragon NaturallySpeaking
#   (c) Copyright 1999 by Joel Gould
#   Portions (c) Copyright 1999 by Dragon Systems, Inc.
#
# _control.py, adapted version of_gramutils.py
# Author: Bart Jan van Os, Version: 1.0, nov 1999
# starting new version Quintijn Hoogenboom, August 2003
"""Grammar that controls/shows/traces state of other grammars


"""
import os
import sys
import re
import pickle

import natlink
from natlinkcore.natlinkutils import *
from dtactions.unimacro import unimacroutils
import unimacro.natlinkutilsbj as natbj
from natlinkcore import natlinkutils
from dtactions.unimacro import unimacroutils
import D_

# extra commands for controlling actions module:
try:
    actions = __import__('actions')
except ImportError:
    actions = None
    print('warning: actions module not imported')

tracecount = list(map(str, list(range(1, 10))))



#Words that are 'filtered out' (actually: removed) in Filter Mode
#See below for the different modes
def ReadFilteredWords(Filename):
#reads all words from the Filtered words file    
#does not really belong here
    try:
        File=open(Filename,'r')
    except:
        return []
    Words = File.readlines()
    File.close
    for w in Words:
        Words[Words.index(w)]=w[:-1]
    freq={}        
    for w in Words:
        if w in freq:
            freq[w]=freq[w]+1
        else:
            freq[w]=1
    Words=list(freq.keys())
    return Words

FilteredWords = ['in','the','minimum','to','and','end','a','of','that','it',
                 'if', 'its', 'is', 'this', 'booth', 'on', 'with',"'s"]
#(taken from natlinkmain, to prevent import:)
baseDirectory = unimacroutils.getUnimacroUserDirectory()
FilterFileName=baseDirectory+'\\filtered.txt'
FilteredWords=natbj.Union(FilteredWords, ReadFilteredWords(FilterFileName))


#Constants for the UtilGrammar
Normal=0
Training=1 #obsolete
Command=2 #obsolete
Filter=4
FilterTraining=5
Display=6


showAll = 1  # reset if no echo of exclusive commands is wished
voicecodeHome = None
if 'VCODE_HOME' in os.environ:
    voicecodeHome = os.environ['VCODE_HOME']
    if os.path.isdir(voicecodeHome):
        for subFolder in ('Admin', 'Config', 'Mediator'):
            newFolder = os.path.join(voicecodeHome, subFolder)
            if os.path.isdir(newFolder) and newFolder not in sys.path:
                sys.path.append(newFolder)
                print('appending to sys.path: %s'% newFolder)
    else:
        print('_control: VCODE_HOME points NOT to a directory: %s'% voicecodeHome)
        voicecodeHome = None
        

ancestor = natbj.IniGrammar
class UtilGrammar(ancestor):
    language = unimacroutils.getLanguage()
    iniIgnoreGrammarLists = ['gramnames', 'tracecount', 'message'] # are set in this module

    name = 'control'
##    normalSet = ['show', 'edit', 'trace', 'switch', 'message']
##    exclusiveSet = ['showexclusive', 'message']
    # commands for controlling module actions
    if actions:
        actionName = "| actions"
    else:
        actionName = ""
        
    gramSpec = """
<show> exported = show ('all grammars' | 
                        ([grammar|inifile] {gramnames}) 
                         """+actionName+""");
<edit> exported = edit ([grammar|inifile] {gramnames}"""+actionName+""");
<trace> exported = trace [on|off| {tracecount}]
              ('all grammars' | [grammar] {gramnames}"""+actionName+""") [on|off| {tracecount}] ;
<switch> exported = switch [on|off]
              ('all grammars' | [grammar] {gramnames})[on|off];
<voicecode> exported = 'switch on voicecode';
<showexclusive> exported = show exclusive [grammars];
<resetexclusive> exported = reset exclusive [grammars];
<message> exported = {message};
    """

    Mode = Normal
    LastMode = Normal
    CurrentWord = 0
    Repeat = 0

    def initialize(self):
        global loadedNames
        if not self.load(self.gramSpec, allResults=1):
            return None
        natbj.RegisterControlObject(self)
        self.emptyList('message')
        self.emptyList('gramnames')
        self.setList('tracecount', tracecount)
        self.activateAll()
        self.setMode(Normal)
        self.doMessages = None
        self.startExclusive = self.exclusive # exclusive state at start of recognition!
##        if unimacroutils.getUser() == 'martijn':
##            print 'martijn, set exclusive %s'% self.name
##            self.setExclusive(1)

    def unload(self):
        natbj.UnRegisterControlObject(self)
        ancestor.unload(self)

    def gotBegin(self, moduleInfo):
        #Now is the time to get the names of the grammar objects and
        # activate the list for the <ShowTrainGrammar> rule
        if natbj.grammarsChanged:
##            print 'new list for control: %s'% natbj.allUnimacroGrammars.keys()
            self.setList('gramnames', list(natbj.allUnimacroGrammars.keys()))
            natbj.ClearGrammarsChangedFlag()
        if self.checkForChanges:
            self.checkInifile()
            
        self.startExclusive = self.exclusive # exclusive state at start of recognition!

    def gotResultsInit(self,words,fullResults):
        if self.mayBeSwitchedOn == 'exclusive':
            self.warning('recog controle, switch off mic: %s', words)
            natbj.SetMic('off')
        if self.exclusive and self.doMessages:
            self.DisplayMessage('<%s>'% ' '.join(words))

    def resetExclusiveMode(self):
        """no activateAll, do nothing, this grammar follows the last unexclusive grammar
        """
        pass
    
    def setExclusiveMode(self):
        """no nothing, control follows other exclusive grammars
        """
        pass
    
    def gotResultsObject(self,recogType,resObj):
        if self.doMessages:
            mes = natbj.GetPendingMessage()
            if mes:
                mes = '\n\n'.join(mes)
                natbj.ClearPendingMessage()
                self.DisplayMessage(mes)
                return
            unimacroutils.doPendingBringUps()  # if there were

        if self.startExclusive and self.exclusive and showAll:
            # display recognition results for exclusive grammars
            # with showAll = None (in top of this file), you can disable this feature
            if recogType == 'reject':
                URName = ''
                exclGram = natbj.exclusiveGrammars
                exclGramKeys = list(exclGram.keys())
                if len(exclGramKeys) > 1 and self.name in exclGramKeys:
                    exclGramKeys.remove(self.name)
                if len(exclGramKeys) > 1:
                    URName = ', '.join(exclGramKeys)
                k = exclGramKeys[0]
                v = exclGram[k]
                # UE stands for unrecognised exclusive:
                # if more exclusive grammars,
                # the last one is taken
                URName = URName or v[2]   # combined name or last rule of excl grammar
                if self.doMessages:
                    self.DisplayMessage('<'+URName+': ???>', alsoPrint=0)

    #Utilities for Filter Modes and other special modes
    def setMode(self,NewMode):
        self.LastMode = self.Mode
        self.Mode = NewMode

    def restoreMode(self):
        self.Mode = self.LastMode
        
    def gotResults_trace(self,words,fullResults):
        self.info('control, trace: %s', words)
        if self.hasCommon(words, 'actions'):
            if self.hasCommon(words, 'show'):
                actions.debugActionsShow()
            elif self.hasCommon(words, 'off'):
                actions.debugActions(0)
            elif self.hasCommon(words, 'on'):
                actions.debugActions(1)
            elif words[-1] in tracecount:
                actions.debugActions(int(words[-1]))
            else:
                actions.debugActions(1)


    def gotResults_voicecode(self,words,fullResults):
        """switch on if requirements are fulfilled

        voicecodeHome must exist
        emacs must be in foreground
        """
        wxmed = os.path.join(voicecodeHome, 'mediator', 'wxmediator.py')
        if os.path.isfile(wxmed):
            commandLine = r"%spython.exe %s > D:\foo1.txt >> D:\foo2.txt"% (sys.prefix, wxmed)
            self.debug("commandLine : %s",commandLine)
            os.system(commandLine)
        else:
            self.info('not a file: %s',wxmed)
            
        

        
    def gotResults_switch(self,words,fullResults):
        self.info('control, switch: %s', words)
        if self.hasCommon(words, 'on'):
            func = 'switchOn'
        elif self.hasCommon(words, 'off'):
            func = 'switchOff'
        else:
            try:
                t = {'nld': '<%s: ongeldig schakel-commando>'% self.GetName()}[self.language]
            except:            
                t = '<%s: invalid switch command>'% self.GetName()
            if self.doMessages:
                self.DisplayMessage(t)
            return
        if self.hasCommon(words, 'all grammars'):
            self.info('%s all grammars:', func)
            natbj.CallAllGrammarObjects(func, ())
            self.info("-"*10)
        else:
            gramname = self.hasCommon(words, list(natbj.allUnimacroGrammars.keys()))
            if gramname:
                gram = natbj.allUnimacroGrammars[gramname]
                gram.callIfExists(func, ())
            else:
                self.warning('no grammar name found: %s'% gramname)
            
    def gotResults_showexclusive(self,words,fullResults):
        if natbj.exclusiveGrammars:
            T = ['exclusive grammars:']
            for e in natbj.exclusiveGrammars:
                T.append(e)
            T.append(self.name)
            if self.doMessages:
                self.DisplayMessage(' '.join(T))
        else:
            if self.doMessages:
                self.DisplayMessage('no exclusive grammars')
            

    def gotResults_resetexclusive(self,words,fullResults):
        print('reset exclusive')
        if natbj.exclusiveGrammars:
            T = ['exclusive grammars:']
            for e in natbj.exclusiveGrammars:
                T.append(e)
            T.append(self.name)
            T.append('... reset exclusive mode')
            if self.doMessages:
                self.DisplayMessage(' '.join(T))
            natbj.GlobalResetExclusiveMode()
            unimacroutils.Wait(1)
            if self.doMessages:
                self.DisplayMessage('reset exclusive mode OK')
        else:
            if self.doMessages:
                self.DisplayMessage('no exclusive grammars')
        
##    def setExclusive(self, state):
##        """control grammar, do NOT register, set and maintain state
##
##        special position because of ControlGrammar
##        """
##        print 'control set exclusive: %s'% state
##        if state == None:
##            return
##        if state == self.exclusive:
##            return
##        print 'control, (re)set exclusive: %s'% state
##        self.gramObj.setExclusive(state)
##        self.exclusive = state

    def gotResults_show(self,words,fullResults):
        # special case for actions:
        if self.hasCommon(words, 'actions'):
            actions.showActions()
            return
        
        if natbj.exclusiveGrammars:
            self.info('exclusive (+ control) are: %s', ' '.join(list(natbj.exclusiveGrammars.keys())))

        grammars = natbj.allUnimacroGrammars
        gramNames = list(grammars.keys())
        gramName = self.hasCommon(words, gramNames)
        if gramName:
            grammar = grammars[gramName]
            if not grammar.onOrOff:
                # off, show message:
                self.offInfo(grammar)
                return
            if not self.hasCommon(words, 'grammar'):
                try:
                    grammar.showInifile()
                    return
                except AttributeError:
                    try:
                        grammar.showGrammar()
                        return
                    except AttributeError:
                        pass
        
        # now show the grammar in the browser application:      
        if gramName:
            name = [gramName]
        else:
            name = words[1:-1]
        
        if (len(name)>0) and (name[0]=='previous'):  #shortcut: show previously assembled grammar
            pypath = '.'
            for x in sys.path:
                pypath = pypath + ';' + x
            os.environ['PYTHONPATH'] = pypath
            AppBringUp('Browser',Exec=PythonwinExe,Args='/app BrowseGrammarApp.py')
        else:
            All=1
            if len(name)>0:
                All=self.hasCommon(words, 'all grammars')
                if self.hasCommon(words, ['all', 'active']):
                    name=name[1:]
            if len(name)>0:                
                Start=(' '.join(name),[])
            else:
                Start=()
            self.info('start browsing with: %s', All)
            self.Browse(Start,All)
        

    def gotResults_edit(self,words,fullResults):
        # special case for actions:
        if self.hasCommon(words, 'actions'):
            actions.editActions()
            return

        grammars = natbj.allUnimacroGrammars
        gramNames = list(grammars.keys())
        gramName = self.hasCommon(words[-1:], gramNames)
        if gramName:
            grammar = grammars[gramName]
            if self.hasCommon(words, 'grammar'):
                module = grammar.__module__
                filename = unimacroutils.getModuleFilename(module)
                #print 'open for edit file: %s'% filename
                self.openFileDefault(filename, mode="edit", name='edit grammar %s'% gramName)
                # unimacroutils.setCheckForGrammarChanges(1)
            else:
                # edit the inifile
                try:
                    grammar.switchOn()
                    grammar.editInifile()
                except AttributeError:
                    if self.doMessages:
                        self.DisplayMessage('grammar "%s" has no method "editInifile"'% gramName)
                    return
        else:
            self.info('no grammar name found')


    def switchOff(self, **kw):
        """overload, this grammar never switches off

        """        
        print('remains switched on: %s' % self)

    def switchOn(self, **kw):
        """overload, just switch on

        """
        self.activateAll()


    def offInfo(self, grammar):
        """gives a nice message that the grammar is switched off

        Gives also information on how to switch on.

        """        
        name = grammar.getName()
        try:
            t = {'nld': ['Grammatica "%s" is uitgeschakeld'% name,
                         '', 
                         '"Schakel grammatica %s in" om te activeren'% name]}[self.language]
            title = {'nld': 'Grammatica %s'% name}[self.language]
        except KeyError:
            t = ['Grammar "%s" is switched off'% name,
                 '"Switch On Grammar %s" to activate'% name]
            title = 'Grammar %s'% name
        if natbj.loadedMessageGrammar:
            t = ';  '.join(t)
            if self.doMessages:
                self.DisplayMessage(t)
        else:
            actions.Message(t)

class MessageDictGrammar(natlinkutils.DictGramBase):
    def __init__(self):
        natlinkutils.DictGramBase.__init__(self)

    def initialize(self):
        self.info('initializing/loading DictGrammar!!')
        self.load()
        natbj.RegisterMessageObject(self)

    def unload(self):
        natbj.UnRegisterMessageObject(self)
        natlinkutils.DictGramBase.unload(self)
        
    def gotResults(self, words):
##        pass
        self.info('messageDictGrammar: heard dictation:  %s ', words)


# standard stuff Joel (adapted for possible empty gramSpec, QH, unimacro)
messageDictGrammar = MessageDictGrammar()
messageDictGrammar.initialize()
print('messageDictGrammar initialized')


# standard stuff Joel (adapted for possible empty gramSpec, QH, unimacro)
utilGrammar = UtilGrammar()
if utilGrammar.gramSpec:
    utilGrammar.initialize()
else:
    print('grammar _control has no specification for this language---------')
    utilGrammar = None

def unload():
    global utilGrammar, messageDictGrammar
    if utilGrammar: utilGrammar.unload()
    utilGrammar = None
    if messageDictGrammar:
        messageDictGrammar.unload()
    messageDictGrammar = None
    
def changeCallback(type,args):
    global utilGrammar
    # Whenever the mic is turned off, the intercept mode is turned off.
    # and any special modes, except training
    if ((type == 'mic') and (args=='on')):
        return   # check WAS in natlinkmain...
    natbj.GlobalResetExclusiveMode()
    if utilGrammar:    
        utilGrammar.setMode(Normal)
        #This could be done anywhere, but not within natlinkutilsbj
        #Because that module is 'imported from'.
        if utilGrammar.interceptMode:
            self.CallAllGrammarObjects('setInterceptMode',[0])
        
        
    



#
# Python Macro Language for Dragon NaturallySpeaking
#   (c) Copyright 1999 by Joel Gould
#   Portions (c) Copyright 1999 by Dragon Systems, Inc.
#
# unittestLilypond.py
#   This script performs some basic tests of the LyNote class which handles lilypond input strings
#   This has nothing to do with speech recognition, but here because a Unimacro grammar "frescobaldi.py"
#   is developed in order to do inputs for lilypond in frescobalde.
#
#pylint:disable=C0209, R0913, R0914

import sys
import unittest
import os
import os.path
from unimacro import lynote
import TestCaseWithHelpers

def getBaseFolder(globalsDict=None):
    """get the folder of the calling module.

    either sys.argv[0] (when run direct) or
    __file__, which can be empty. In that case take the working directory
    """
    globalsDictHere = globalsDict or globals()
    baseFolder = ""
    if globalsDictHere['__name__']  == "__main__":
        baseFolder = os.path.split(sys.argv[0])[0]
        print('baseFolder from argv: %s'% baseFolder)
    elif globalsDictHere['__file__']:
        baseFolder = os.path.split(globalsDictHere['__file__'])[0]
        print('baseFolder from __file__: %s'% baseFolder)
    if not baseFolder or baseFolder == '.':
        baseFolder = os.getcwd()
        print('baseFolder was empty, take wd: %s'% baseFolder)
    return baseFolder

thisDir = getBaseFolder(globals())
logFileName = os.path.join(thisDir, "testresult.txt")

#---------------------------------------------------------------------------
# These tests should be run after we call natConnect
# no reopen user at each test anymore..
# no default open window (open window will be the calling program)
# default .ini files pop up when you first run this test. just ignore them.
class UnittestLyNote(TestCaseWithHelpers.TestCaseWithHelpers):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def checkLyResult(self, expNote, lyn):
        """check the parts of the note (lyn) against the expected values
        """
        note, elevation, duration, additions, backslashed = expNote
        self.assertEqual(note, lyn.note, 'note not equal "%s", expected: "%s", got: "%s"'% (lyn.originalInput, note, lyn.note))
        self.assertEqual(elevation, lyn.elevation, 'elevation not equal "%s", expected: "%s", got: "%s"'% (lyn.originalInput, elevation, lyn.elevation))
        self.assertEqual(duration, lyn.duration, 'duration not equal "%s", expected: "%s", got: "%s"'% (lyn.originalInput, duration, lyn.duration))
        self.assertEqual(additions, lyn.additions, 'additions not equal "%s", expected: "%s", got: "%s"'% (lyn.originalInput, additions, lyn.additions))
        self.assertEqual(backslashed, lyn.backslashed, 'backslashed not equal "%s", expected: "%s", got: "%s"'% (lyn.originalInput, backslashed, lyn.backslashed))
    
    def testSimpleNote(self):
        """test a few basic examples of a lilypond note
        """
        for s, expNote in [("bf,8.", ("bf", ",", "8.", "", None)),
                        ("g,8.", ("g", ",", "8.", "", None)),
                        ("a,,8", ("a", ",,", "8", "", None)),
                        ("cis", ("cis", "", "", "", None)),
                        ("aes'(", ("aes", "'", "", "(", None)),
                        (r"d4([\melisma", ("d", "", "4", "([", ['\\melisma'])),
                        (r"ais,,32..\melismaEnd]\break\)", ("ais", ",,", "32..", r"]\)", [r'\melismaEnd', r'\break']))]:
                        
            lyn = lynote.LyNote(s)
            self.checkLyResult( expNote, lyn )
            self.assertTrue(lyn.isNote(), '"%s" should be a note (isNote function), not: %s'% (s, lyn.isNote()))
            print('testSimpleNote OK: "%s", result: "%s"'% (s, lyn))

    def testIncompleteNoteAndUpdate(self):
        """test a few basic examples of a lilypond note
        """
        for s, expNote in [(",8.", ("", ",", "8.", "", None)),
                        ("8", ("", "", "8", "", None)),
                        (r"\melismaEnd", ("", "", "", "", [r"\melismaEnd"])),
                        (r",,1\melismaEnd]\break\)", ("", ",,", "1", r"]\)", [r'\melismaEnd', r'\break']))]:
                        
            lyn = lynote.LyNote(s)
            self.checkLyResult( expNote, lyn )
            print('testIncompleteNote OK: "%s", result: "%s"'% (s, lyn))

        updateNote = "4"            
        orgNote = "bf"
        lyorg = lynote.LyNote(orgNote)
        lyorg.updateNote(updateNote)
        expUpdated = ("bf", "", "4", "",  None)
        self.checkLyResult( expUpdated, lyorg )
        print('testIncompleteNote "%s" updated with "%s" OK, result: "%s"'% (orgNote, updateNote, lyorg))



        updateNote = ","            
        orgNote = "g'4("
        lyorg = lynote.LyNote(orgNote)
        lyorg.updateNote(updateNote)
        expUpdated = ("g", "", "4", "(", None)
        self.checkLyResult( expUpdated, lyorg )
        print('testIncompleteNote "%s" updated with "%s" OK, result: "%s"'% (orgNote, updateNote, lyorg))

        updateNote = "8"            
        orgNote = "a'4("
        lyorg = lynote.LyNote(orgNote)
        lyorg.updateNote(updateNote)
        expUpdated = ("a", "'", "8", "(", None)
        self.checkLyResult( expUpdated, lyorg )
        print('testIncompleteNote "%s" updated with string "%s" OK, result: "%s"'% (orgNote, updateNote, lyorg))

        lyorg = lynote.LyNote(orgNote)
        lyupdate = lynote.LyNote(updateNote)
        lyorg.updateNote(lyupdate)
        self.checkLyResult( expUpdated, lyorg )
        print('testIncompleteNote "%s" updated with instance "%s" OK, result: "%s"'% (orgNote, lyupdate, lyorg))






        updateNote = r"c'8..\melisma"            
        orgNote = "b'4("
        lyorg = lynote.LyNote(orgNote)
        lyorg.updateNote(updateNote)
        expUpdated = ("c", "''", "8..", "(", [r"\melisma"])
        self.checkLyResult( expUpdated, lyorg )
        print('testIncompleteNote "%s" updated with "%s" OK, result: "%s"'% (orgNote, updateNote, lyorg))

        lyorg = lynote.LyNote(orgNote)
        lyupdate = lynote.LyNote(updateNote)
        lyorg.updateNote(lyupdate)
        self.checkLyResult( expUpdated, lyorg )
        print('testIncompleteNote "%s" updated with instance "%s" OK, result: "%s"'% (orgNote, lyupdate, lyorg))

    def testReBackslashedWord(self):
        r"""test \break, \melisma etc, but refuse \( and \)"""
        for s in ['', r'\)']:
            m = lynote.reBackslashedWord.match(s)
            self.assertEqual(None, m, 'string "%s" should not match reBackslashWord'% s)
            print('testReBackslashedWord: correct fail to match "%s"'% s)
        for s in [r'\break', r'\melismaEnd']:
            m = lynote.reBackslashedWord.match(s)
            if m is None:
                self.fail('string r"%s" should match reBackslashWord'% s)
            print('testReBackslashedWord: correct match of "%s"'% s)
            
    def testStrAndRepr(self):
        """check the correct str and repr result"""        
        for s in ["g,8.", "a,,8"]:
            expStr = s
            lyn = lynote.LyNote(s)
            self.assertEqual(expStr, str(lyn), 'str of "%s" fails ("%s")'% (s, str(lyn)))
            print('testStrAndRepr: correct str "%s" (being the same)'% s)
            expRepr = "lynote: " + s
            self.assertEqual(expRepr, repr(lyn), 'repr of "%s" fails ("%s")'% (s, repr(lyn)))
            print('testStrAndRepr: correct repr "%s" (of "%s")'% (s, repr(s)))
    
    def testAnalyseString(self):
        """check the analyseString function"""
        
        expWords = ['a4', ' ', 'b', ' ', 'c']
        expNotesIndexes =  [0, 2, 4]
        s = 'a4 b c'
        words, noteIndexes = lynote.analyseString(s)
        words = [str(n) for n in words]
        self.assertEqual(expWords, words, 'words of testAnalyseString (%s) not as expected\nexpected: %s\ngot: %s'%
                         (s, expWords, words))
        self.assertEqual(expNotesIndexes, noteIndexes, 'noteIndexes of testAnalyseString (%s) not as expected\nexpected: %s\ngot: %s'%
                         (s, expNotesIndexes, noteIndexes))

        # testing the beforeWord case:
        s = ' | b8\n'
        expWords= [' | ', 'b8', '\n']
        expNotesIndexes =  [1]
        words, noteIndexes = lynote.analyseString(s)
        words = [str(n) for n in words]
        self.assertEqual(expWords, words, 'words of testAnalyseString (%s) not as expected\nexpected: %s\ngot: %s'%
                         (s, expWords, words))
        self.assertEqual(expNotesIndexes, noteIndexes, 'noteIndexes of testAnalyseString (%s) not as expected\nexpected: %s\ngot: %s'%
                         (s, expNotesIndexes, noteIndexes))

        # testing the beforeWord case:
        s = 'c4\n\n\n\n | b'
        expWords= ['c4', '\n\n\n\n | ', 'b']
        expNotesIndexes =  [0, 2]
        words, noteIndexes = lynote.analyseString(s)
        words = [str(n) for n in words]
        self.assertEqual(expWords, words, 'words of testAnalyseString (%s) not as expected\nexpected: %s\ngot: %s'%
                         (s, expWords, words))
        self.assertEqual(expNotesIndexes, noteIndexes, 'noteIndexes of testAnalyseString (%s) not as expected\nexpected: %s\ngot: %s'%
                         (s, expNotesIndexes, noteIndexes))





        # lot of text after last note:
        expWords = ['ais', '  ', 'b8', '\n\\time 3/4\n', 'g4', ' ', 'a', '  ', 'b', ' ', 'c', ' \\bar "||"\n', 'c1']
        expNotesIndexes =   [0, 2, 4, 6, 8, 10, 12]
        s = 'ais  b8\n\\time 3/4\ng4 a  b c \\bar "||"\nc1'
        words, noteIndexes = lynote.analyseString(s)
        words = [str(n) for n in words]
        self.assertEqual(expWords, words, 'words of testAnalyseString (%s) not as expected\nexpected: %s\ngot: %s'%
                         (s, expWords, words))
        self.assertEqual(expNotesIndexes, noteIndexes, 'noteIndexes of testAnalyseString (%s) not as expected\nexpected: %s\ngot: %s'%
                         (s, expNotesIndexes, noteIndexes))

    def doTestGetNotesFromWordsList(self, maxWords, wordsList, notesIndexes,
             expWords, expNotesIndexes, expBeforeNotes, expAfterNotes, expAfterNoteComplete, reverse=False):
        """test the variants for the getting of notes from a wordslist
        
        normal, for next notes and reverse, for previous notes.
        
        is called from routines below. several notes are tested in one run
        
        """
        for expName in 'expWords', 'expNotesIndexes', 'expBeforeNotes', 'expAfterNotes', 'expAfterNoteComplete':
            exp = locals()[expName]
            self.assertEqual(maxWords, len(exp), 'doTestGetNotesFromWordsList, %s should be sequence of length %s, not %s\nwordsList: %s'%
                (expName, maxWords, len(expWords), repr(wordsList)))
            
        if reverse:
            # setting expBeforeNoteComplete:
            expBeforeNoteComplete = expAfterNoteComplete
            
        
        for i in range(maxWords):
            numWords = i + 1
            print('---testing %s words from %s'% (numWords, wordsList))
        
            if reverse:
                # tricky with calling names:        
                gotWordsList, gotNotesIndexes, afterNotes, beforeNotes, beforeNoteComplete = \
                        lynote.getNotesFromWordsList(numWords, wordsList, notesIndexes, reverse=reverse)
            else:        
                gotWordsList, gotNotesIndexes, beforeNotes, afterNotes, afterNoteComplete = \
                        lynote.getNotesFromWordsList(numWords, wordsList, notesIndexes)
            gotWords = [str(a) for a in gotWordsList]
        
            self.assertEqual(expWords[i], gotWords, 'words of getNotesFromWordsList (%s) not as expected\nexpected: %s\ngot: %s'%
                         (numWords, expWords[i], gotWords))
            self.assertEqual(expNotesIndexes[i], gotNotesIndexes, 'noteIndexes of getNotesFromWordsList (%s) not as expected\nexpected: %s\ngot: %s'%
                         (numWords, expNotesIndexes[i], gotNotesIndexes))
            if reverse:
                # for previous words:
                self.assertEqual(expBeforeNotes[i], beforeNotes, 'REVERSE: beforeNotes of getNotesFromWordsList (%s) not as expected\nexpected: %s\ngot: %s'%
                             (numWords, repr(expBeforeNotes[i]), repr(beforeNotes)))
                self.assertEqual(expAfterNotes[i], afterNotes, 'REVERSE: afterNotes of getNotesFromWordsList (%s) not as expected\nexpected: %s\ngot: %s'%
                             (numWords, repr(expAfterNotes[i]), repr(afterNotes)))
                self.assertEqual(expBeforeNoteComplete[i], beforeNoteComplete, 'REVERSE beforeNoteCompleteof getNotesFromWordsList (%s) not as expected\nexpected: %s\ngot: %s'%
                             (numWords, expBeforeNoteComplete[i], beforeNoteComplete))
            else:
                # normal case, next words:
                self.assertEqual(expBeforeNotes[i], beforeNotes, 'beforeNotes of getNotesFromWordsList (%s) not as expected\nexpected: %s\ngot: %s'%
                             (numWords, repr(expBeforeNotes[i]), repr(beforeNotes)))
                self.assertEqual(expAfterNotes[i], afterNotes, 'afterNotes of getNotesFromWordsList (%s) not as expected\nexpected: %s\ngot: %s'%
                             (numWords, repr(expAfterNotes[i]), repr(afterNotes)))
                self.assertEqual(expAfterNoteComplete[i], afterNoteComplete, 'afterNoteCompleteof getNotesFromWordsList (%s) not as expected\nexpected: %s\ngot: %s'%
                             (numWords, expAfterNoteComplete[i], afterNoteComplete))

                

    def tttestGetNotesFromWordsList(self):
        """check the getNotesFromWordsList function
        
        here are the "forward" next note tests, with reverse == False (default)
        See the function getNotesFromWordsList in lynote.py
        """

        # testing the beforeWord case:
        s = ' | b8\n'
        expWords= [' | ', 'b8', '\n']
        expNotesIndexes =  [1]
        # first redo the analyseString test:
        wordsList, notesIndexes = lynote.analyseString(s)
        gotWordsList = [str(a) for a in wordsList]
        self.assertEqual(expWords, gotWordsList, 'assertion test: words of analyseString (%s) not as expected\nexpected: %s\ngot: %s'%
                         (s, expWords, gotWordsList))
        self.assertEqual(expNotesIndexes, notesIndexes, 'assertion test: notesIndexes of testAnalyseString (%s) not as expected\nexpected: %s\ngot: %s'%
                         (s, expNotesIndexes, notesIndexes))
        # now the real test
        maxWords = 1
        expWords =(['b8'],)
        expNotesIndexes = ([0],)
        expBeforeNotes = (' | ',)
        expAfterNotes = ('\n',)
        expAfterNoteComplete = (False,)
        self.doTestGetNotesFromWordsList(maxWords, wordsList, notesIndexes,
             expWords, expNotesIndexes, expBeforeNotes, expAfterNotes, expAfterNoteComplete)
        
        expWords= ['a4', ' ', 'b', ' ', 'c']
        expNotesIndexes =  [0, 2, 4]
        s = 'a4 b c'
        # first redo the analyseString test:
        wordsList, notesIndexes = lynote.analyseString(s)
        gotWordsList = [str(a) for a in wordsList]
        self.assertEqual(expWords, gotWordsList, 'assertion test: words of analyseString (%s) not as expected\nexpected: %s\ngot: %s'%
                         (s, expWords, gotWordsList))
        self.assertEqual(expNotesIndexes, notesIndexes, 'assertion test: notesIndexes of testAnalyseString (%s) not as expected\nexpected: %s\ngot: %s'%
                         (s, expNotesIndexes, notesIndexes))
        # now the real test
        maxWords = 3
        expWords = ['a4'], ['a4', ' ', 'b'], ['a4', ' ', 'b', ' ', 'c']
        expNotesIndexes = [0], [0,2], [0,2,4]
        expBeforeNotes = '', '', ''
        expAfterNotes = ' ', ' ', ''
        expAfterNoteComplete = True, True, False
        self.doTestGetNotesFromWordsList(maxWords, wordsList, notesIndexes,
             expWords, expNotesIndexes, expBeforeNotes, expAfterNotes, expAfterNoteComplete)
        
        # lot of text after last note:
        expWords = ['f', '  ', 'b8', '\n\\time 3/4\n', 'g4', ' ', 'a', '  ', 'b', ' ', 'c', ' \\bar "||"\n', 'c1']
        expNotesIndexes =  [0, 2, 4, 6, 8, 10, 12]
        s = 'f  b8\n\\time 3/4\ng4 a  b c \\bar "||"\nc1'

        wordsList, notesIndexes = lynote.analyseString(s)
        gotWordsList = [str(n) for n in wordsList]
        self.assertEqual(expWords, gotWordsList, 'assertion test: wordsList of testAnalyseString (%s) not as expected\nexpected: %s\ngot: %s'%
                         (s, expWords, gotWordsList))
        self.assertEqual(expNotesIndexes, notesIndexes, 'assertion test: notesIndexes of testAnalyseString (%s) not as expected\nexpected: %s\ngot: %s'%
                         (s, expNotesIndexes, notesIndexes))

        maxWords = 4
        expWords = (['f'], ['f', '  ', 'b8'],
                        ['f', '  ', 'b8', '\n\\time 3/4\n', 'g4'],
                        ['f', '  ', 'b8', '\n\\time 3/4\n', 'g4', ' ', 'a'])
        expNotesIndexes = [0], [0,2], [0,2,4], [0, 2, 4, 6]
        expBeforeNotes = '', '', '', ''
        expAfterNotes = '  ', '\n\\time 3/4\n', ' ', '  '
        expAfterNoteComplete = True, True, True, True
        self.doTestGetNotesFromWordsList(maxWords, wordsList, notesIndexes,
             expWords, expNotesIndexes, expBeforeNotes, expAfterNotes, expAfterNoteComplete)
        
    def tttestGetNotesFromWordsListReverse(self):
        """check the getNotesFromWordsList function

        here are the "backward" PREVIOUS note tests, with reverse == True
        See the function getNotesFromWordsList in lynote.py
        """
        # testing the beforeWord case:
        # s = ' | b8\n'
        # expWords= [' ', '|', ' ', 'b8', '\n']
        # expNotesIndexes =  [3]
        # # first redo the analyseString test:
        # wordsList, notesIndexes = lynote.analyseString(s)
        # gotWordsList = [str(a) for a in wordsList]
        # self.assertEqual(expWords, gotWordsList, 'assertion test: words of analyseString (%s) not as expected\nexpected: %s\ngot: %s'%
        #                  (s, expWords, gotWordsList))
        # self.assertEqual(expNotesIndexes, notesIndexes, 'assertion test: notesIndexes of testAnalyseString (%s) not as expected\nexpected: %s\ngot: %s'%
        #                  (s, expNotesIndexes, notesIndexes))
        # # now the real test
        # maxWords = 1
        # expWords =(['b8'],)
        # expNotesIndexes = ([0],)
        # expBeforeNotes = (' | ',)
        # expAfterNotes = ('\n',)
        # expAfterNoteComplete = (False,)
        # self.doTestGetNotesFromWordsList(maxWords, wordsList, notesIndexes,
        #      expWords, expNotesIndexes, expBeforeNotes, expAfterNotes, expAfterNoteComplete, reverse=True)
        # 
        
        # previous words test:
        expWords= ['a4', ' |\n', 'b', ' ', 'c', ' ']
        expNotesIndexes =  [0, 2, 4]
        s = 'a4 |\nb c '
        # first redo the analyseString test:
        wordsList, notesIndexes = lynote.analyseString(s)
        gotWordsList = [str(a) for a in wordsList]
        self.assertEqual(expWords, gotWordsList, 'assertion test: words of analyseString (%s) not as expected\nexpected: %s\ngot: %s'%
                         (s, expWords, gotWordsList))
        self.assertEqual(expNotesIndexes, notesIndexes, 'assertion test: notesIndexes of testAnalyseString (%s) not as expected\nexpected: %s\ngot: %s'%
                         (s, expNotesIndexes, notesIndexes))
        # now the real test
        maxWords = 3
        expWords = ['c'], ['b', ' ', 'c'], ['a4', ' |\n', 'b', ' ', 'c']
        expNotesIndexes = [0], [0,2], [0,2,4]
        expBeforeNotes = ' ', ' |\n', ''
        expAfterNotes = ' ', ' ', ' '
        expAfterNoteComplete = True, True, False
        self.doTestGetNotesFromWordsList(maxWords, wordsList, notesIndexes,
             expWords, expNotesIndexes, expBeforeNotes, expAfterNotes, expAfterNoteComplete, reverse=True)
      
        
            
def log(t):
    print(t)

def run():
    log('starting unittestLyNote')
    # trick: if you only want one or two tests to perform, change
    # the test names to her example def tttest....
    # and change the word 'test' into 'tttest'...
    # do not forget to change back and do all the tests when you are done.
    suite = unittest.makeSuite(UnittestLyNote, 'test')
##    natconnectOption = 0 # no threading has most chances to pass...
    result = unittest.TextTestRunner().run(suite)
    print(result)
if __name__ == "__main__":
    run()



# http://unimacro.SourceForge.net and http://qh.antenna.nl/unimacro
# (c) copyright 2003 see http://qh.antenna.nl/unimacro/aboutunimacro.html
#    or the file COPYRIGHT.txt in the natlink\natlink directory 
#
#  _tasks.py
#
#
# written by Quintijn Hoogenboom (QH softwaretraining & advies),
# july 2006
# reworked for python3 release: September 2022
#
#pylint:disable = C0302, R0904, C0209, R0914,, R0911, R0912, R0913
#pylint:disable=E1101

"""Do task switching (taskbar), 

And switch to Documents inside an application.

Move tasks to other display other to a corner of the current monitor
"""
#
#

import sys
import time
import os
import os.path
import win32gui

import natlink
from natlinkcore import nsformat
from natlinkcore import natlinkstatus
from unimacro import natlinkutilsbj as natbj
from dtactions.unimacro import unimacroutils 
from dtactions import monitorfunctions
from dtactions.unimacro.unimacroactions import doAction as action
from dtactions.unimacro.unimacroactions import doKeystroke as keystroke
from dtactions.unimacro.unimacroactions import setPosition
from dtactions.unimacro.unimacroactions import getPosition
from dtactions.unimacro import unimacroactions as actions
import win32con
import win32api

status = natlinkstatus.NatlinkStatus()
language = status.language        
ICAlphabet = natbj.getICAlphabet(language=language)

# center mouse after taskswitch (not good with XP and choice boxes in taskbar)

ancestor=natbj.DocstringGrammar
class ThisGrammar(ancestor):
    iniIgnoreGrammarLists = ['taskcount', 'documentcount',
                             #'switchapp', 'degrees', 'pixelcount',
                             'switchapp',
                             #'sizecount','percentcount', 'directionplus', 'direction',
                             'namedtask']
    name = "tasks"
    # task commands in docstring form in the rule functions below
    gramSpec = ["""

# miscelaneous:
<fileswitch> exported  = ("switch file to") {switchapp};
    """]
    
    # include only if searchinothertask is True (via .ini file) (enable search commands = T)
    searchRule = "<searchinothertask> exported = search ({taskcount}|{application}) | <before> search ({taskcount}|{application});"

    def __init__(self):
        """start the inifile and add to grammar if needed the searchRule
        """
        ancestor.__init__(self)
        self.startInifile()
        #print 'enableSearchCommands: %s'% self.enableSearchCommands
        if self.enableSearchCommands:
            self.gramSpec += '\n' + self.searchRule
        _name = self.getName()


    def initialize(self):
        self.load(self.gramSpec)
        self.switchOnOrOff() # initialises lists from inifile, and switches on
                         # if all goes well (and variable onOrOff == 1)
        self.taskCounts = list(range(1, self.maxTaskNumber+1))
        self.documentCounts = list(range(1, self.maxDocumentNumber+1))
        # call for extra spoken form with difficult numbers
        # needs self.stripSpokenForm in gotResults... function!!!
        #self.addSpokenFormNumbers(self.taskCounts)
        self.setNumbersList('taskcount', self.taskCounts)
        self.setNumbersList('documentcount', self.documentCounts)
        self.setList('switchapp', self.switchApps)
        self.emptyList('namedtask')  # to be filled with task name command
        #for moving the window:
        #self.pixelCounts = range(1, 21) + range(20, 201, 10)
        #self.setNumbersList('pixelcount', self.pixelCounts)
        #self.degreeCounts = range(0, 361,10)
        #self.setNumbersList('degrees', self.degreeCounts)
        #self.sizeCounts = range(1,21)
        #self.setNumbersList('sizecount', self.sizeCounts)
        #self.percentCounts = range(10,101,10)
        #self.setNumbersList('percentcount', self.percentCounts)
        self.namedtaskDict = {} # name -> hndle dict
        # the spoken forms are the values of the inifile
        # the keys of the inifile (fixed) are the resulting directions to be
        # worked with
        # for the task position etc commands:
        #left|up|right|down|lefttop|righttop|rightbottom|leftbottom;  
        # self.directionsplus = self.iniGetInvertedDict('directionplusreverse')
        # QH TODO, repair the direction things:
        # self.setList('directionplus', list(self.directionsplus.keys()))
        self.winkeyDown = 0  # for keeping down in stacked taskbar items
        self.switchOnOrOff() # initialises lists from inifile, and switches on# if all goes well (and variable onOrOff == 1)
        print('IniGrammar tasks, all lists initialized...')

    def iniGetInvertedDict(self, section):
        """get all the values as keys, and the keys as values
            values are separated by ; to are read in as a list
        """
        #pylint:disable=W0212
        
        D = {}
        keys = self.ini.get(section)
        if not keys:
            raise ValueError('Grammar %s: no keys in inifile section [%s] (inifile: %s)'%
                             (self.name, section, self.ini._name))
        for k in keys:
            values = self.ini.getList(section, k)
            if values:
                for v in values:
                    if v:
                        D[v] = k
            else:
                D[k] = k
        return D

    def gotBegin(self,moduleInfo):
        if self.checkForChanges:
            self.checkInifile() # refills grammar lists and instance variables
        if self.winkeyDown:
            className = unimacroutils.getClassName()
            if className != 'TaskListThumbnailWnd':
                print('tasks, call cancelmode from gotBegin')
                self.cancelMode()
        if self.makeTaskNamesPythonw and moduleInfo and moduleInfo[0].endswith("pythonw.exe"):
            wTitle = moduleInfo[1]
            if wTitle.find(" --- ") > 0:
                self.addTaskNameToDict(moduleInfo)

    def gotResultsInit(self, words, fullResults):
        self.fullResults = fullResults 
        # for the move commands:
        self.directionplus = None # for position, stretch, shrink
        self.units = None
        self.amount = None
        self.taskmoveresize = None  # also "position" (see gotResults)
        self.hadStartMenu = 0
        self.letters = None
        self.app = None
        self.namedictated = ""  # for naming a window
        self.giveName = None # for the hndle when naming a window
    ### here come the docstring like rules:
    ###

    def rule_startmenu(self, words):
        """start menu [{startmenucommands}]
        """
        print('got start menu')
        if self.hasCommon(words[0], 'start'):
            self.hadStartMenu = 1
        Act = self.getFromInifile(words[-1], 'startmenucommands')
        if Act:
            self.hadStartMenu = 0 # no action in gotResults, doing it here
            self.doStartMenu()
            action(Act)
    
    def doStartMenu(self):
        """give the start menu
        """
        action("SSK {ctrl+esc}")
        
    def rule_taskswitch(self, words):
        """#commands for switching tasks:    
        task ({taskcount}|{taskcount}<subtask>|{application}|{application}<subtask>|back|{namedtask})
        """
        #print 'got taskswitch: %s'% words
        #switch to a task by number, application or Back
        #leave the mouse at 0.3 0.3 relative to the top
        countOrApp = words[1]
        #print 'goto task, countOrApp: %s'% countOrApp
        result = self.gotoTask(countOrApp)
        
        if result:
            prog_info = unimacroutils.getProgInfo()
            if prog_info.prog == 'explorer' and not prog_info.title:
                return # no centermouse!
            if self.centerMouse and not self.nextRule:
                unimacroutils.Wait()
                unimacroutils.doMouse(1, 5, 0.3, 0.3, 0, 0)  # relative in client area, no clicking           
        else:
            print('_tasks, could not switch to task: %s'% countOrApp)


    def rule_subtask(self, words):
        """# commands for going to a subtask in a stacked taskbar
        window {n1-10} | {firstlast} window
        """
        className = unimacroutils.getClassName()
        wNumList = self.getNumbersFromSpoken(words) # returns a string or None
        if wNumList:
            wNum = wNumList[0]
        else:
            wNum = self.getFromInifile(words, 'firstlast')
        if not wNum:
            print('tasks, rule subtaks, no window number found')
            return
        
        #print 'tasks, subtask: %s'% wNum
        if className == "TaskListThumbnailWnd":
            #print 'got subtask, alternative window: %s'% wNum
            self.doAlternativeClick(className, wNum)

        if self.centerMouse and not self.nextRule:  # so last rule of the recognition
            unimacroutils.doMouse(1, 5, 0.3, 0.3, 0, 0)

    def rule_numbereddocument(self, words):
        """# go to a numbered document
        document ({documentcount} [{documentaction}] | {documentaction})
        """
        count = self.getNumberFromSpoken(words[1]) # returns a string or None
        if count:
            #print 'goto task, countOrApp: %s'% countOrApp
            result = self.gotoDocument(count)
            
            if result:
                if self.centerMouse:
                    unimacroutils.Wait()
                    unimacroutils.doMouse(1, 5, 0.3, 0.3, 0, 0)  # relative in client area, no clicking           
            else:
                prog_info = unimacroutils.getProgInfo()
                print('_tasks, could not switch to document: %s (program: %s)'% (count, prog_info.prog))
            
            if words[1] == words[-1]:
                return

        Act = self.getFromInifile(words[-1], 'documentaction')
        print('words[-1]: %s, Action: %s'% (words[-1], Act))
        if Act:
            action(Act)

    def rule_relativedocument(self, words):
        """# go to a relative document (previous, next, also with counts)
        (next|previous) document [{n2-10}] [{documentaction}]
        """
        hadNext = self.hasCommon(words[0], 'next')
        count = self.getNumberFromSpoken(words) or 1
        if hadNext:
            for _ in range(count):
                action("<<documentnext>>")
        else:
            for _ in range(count):
                action("<<documentprevious>>")
            
        if words[1] == words[-1]:
            # no document action
            return

        Act = self.getFromInifile(words[-1], 'documentaction')
        print('words[-1]: %s, Action: %s'% (words[-1], Act))
        if Act:
            action(Act)

    def rule_relativedocumentsaction(self, words):
        """# do action on more (relative) documents (previous, next, also with counts)
        [(next|previous)] {n1-20} documents {documentaction}
        """
        nextPrevAction = ""
        hadPrevious = ""
        hadNext = self.hasCommon(words[0], 'next')
        if hadNext:
            nextPrevAction = "<<documentnext>>"
        else:
            hadPrevious = self.hasCommon(words[0], 'previous')
            if hadPrevious:
                nextPrevAction = "<<documentprevious>>"
            
        count = self.getNumberFromSpoken(words[0:2]) # returns a string or None
        if not count:
            print('_tasks, relativedocuments, count expected, got %s'% words[1])

        if hadNext:
            print('next, go one doc forward')
            action(nextPrevAction)
        else:
            print('previous, go %s docs back'% count)
            ## cycle n times:
            for _ in range(count):
                action(nextPrevAction)

        Act = self.getFromInifile(words[-1], 'documentaction')
        if not Act:
            print('_tasks, relativedocuments, action expected, got %s'% words)
            return
        for _ in range(count):
            print('do action  %s (%s times)'% (Act, count))
            action(Act)

    def gotResults_searchinothertask(self, words, fullResults):
        ## rule for searching in another task, activated only if ...
        
        t = time.time()
        searchWord = self.getSelectedText()
        print('searchWord from getSelectedText: %s'% searchWord)
        if not searchWord:
            searchWord = action("SELECTWORD")
            if not searchWord:
                print('_tasks, searchinothertask: could not select text')
                return
        t1 = time.time() 
        print('searchword: %s (%.2f)'% (searchWord, t1-t))
        countOrApp = words[1]
        #print 'switch to task: %s'% countOrApp
        result = self.gotoTask(countOrApp)
        if result is None:
            print('_tasks, could not switch to task: %s'% countOrApp)
            return
        print('result after taskswitch: %s'% repr(result))
        # t2 = time.time() 

        prog_info = result
        #print 'switched to "%s" (%.2f)'%  (prog, t2-t1)


        if prog_info.prog == 'explorer' and not prog_info.title:
            return # no centermouse!
        if self.centerMouse:
            unimacroutils.Wait()
            unimacroutils.doMouse(1, 5, 0.3, 0.3, 0, 0)  # relative in client area, no clicking           

        unimacroutils.Wait()
        # now do the postprocessing
        prog = prog_info.prog
        print('postprocessing for search: %s in app: %s'% (searchWord, prog))
        if prog in ['chrome', 'firefox','iexplore']:
            phrase = '"%s" site:nl'% searchWord
            keystroke("{ctrl+k}")
            keystroke(phrase + "{enter}")
        elif prog_info.prog == 'dictionaryworkbench':
            # hardcoded for Arnoud
            keystroke('{ctrl+f}')
            action("SCLIP %s"% searchWord)
            keystroke('{enter}')
        else:
            #t3 = time.time()
            action("<<startsearch>>", progInfo=prog_info)
            #t4 = time.time()
            keystroke(searchWord, progInfo=prog_info)
            #t5 = time.time()
            action("<<searchgo>>", progInfo=prog_info)
            #t6 = time.time()
            #print 'after searchaction: %s (%.2f, %.2f, %.2f, %.2f)'% (searchWord, t3-t2,
            #                                                    t4-t3, t5-t4, t6-t5)
            

    def subrule_before(self, words):
        """#optional word here in front of command:
        Here
        """
        action("CLICK")
        unimacroutils.visibleWait()
        #action("CLICKIFSTEADY")

    def importedrule_dgndictation(self, words):
        # for giving a window a name
        self.namedictated = nsformat.formatWords(words, state=-1)[0]
        

    def rule_taskaction(self, words):
        '(<taskswitch> | <subtask> | task)({taskaction}| "give name" <dgndictation>|<closemultiple>)'
        
        # do a taskaction to the current task, or one of the
        # numbered/named tasks or task back
        #print 'got taskaction: %s'% words
        Actions = self.ini.get('taskaction')
        act = self.hasCommon(words, Actions)
        giveName = self.hasCommon(words, "give name")
        #if self.app == "messages":
        #    print 'action for messages: %s'% act
        #    if act == 'refresh':
        #        self.doTaskAction('close')
        #        print 'new messages window'
        #        return
        if giveName:
            hndle = natlink.getCurrentModule()[2]
            self.giveName = hndle
        elif act:
            self.doTaskAction(act)
        else:
            print('thistask in _general, no valid action', words)

        prog_info = unimacroutils.getProgInfo()
        if prog_info.prog == 'explorer' and not prog_info.title:
            return # no centermouse!
        
        if self.centerMouse:
            unimacroutils.doMouse(1, 5, 0.3, 0.3, 0, 0)
        
    def subrule_closemultiple(self, words):
        """close (all|other) | (all|other) close
        """
        # note the grammar alternatives, eg in Dutch the translation of
        # all close (alles sluiten) is more convenient.
        if not self.lastTaskCount:
            print('_tasks, close all | close multiple only works with a numbered task')
            return
        _all = self.hasCommon(words, "all")
        multiple = self.hasCommon(words, "multiple")
        if _all:
            action("MP 2, 0, 0, right; VW; {up}{enter}; VW")
            self.gotoTask("back")
            
        elif multiple:
            # close as long as special stacked window is found
            mousePos = unimacroutils.getMousePosition()
            if mousePos is None:
                raise ValueError("could not get mouse position")
            _x, _y = mousePos
            className = unimacroutils.getClassName()
            wNum = -1 # last window of stacked windows...
            #print 'tasks, subtask: %s'% wNum
            while className == "TaskListThumbnailWnd":
                #print 'got subtask, alternative window: %s'% words
                self.doAlternativeClick(className, wNum)
                action("<<windowclose>>")
                action("VW; MP 0, %s, %s"% mousePos)
                className = unimacroutils.getClassName()
            self.gotoTask(self.lastTaskCount)
        if self.centerMouse:
            unimacroutils.doMouse(1, 5, 0.3, 0.3, 0, 0)
            
        
    def gotoTask(self, countOrApp):
        """go to the specified task, by number or application name, return proginfo, or None if task was not found
        
        """
        self.lastTaskCount = None
        if isinstance(countOrApp, str):
            countBack = self.getNumberFromSpoken(countOrApp, self.taskCounts) # returns a string or None
        elif isinstance(countOrApp, int):
            countBack = countOrApp
        appList = self.ini.get('application')
##        print 'appList: %s'% appList
        if isinstance(countOrApp, str) and self.hasCommon(countOrApp, 'back'):
            action('SSK {Alt+Tab}')
            return 1
        if countOrApp in self.namedtaskDict:
            hndle = self.namedtaskDict[countOrApp]
            result = unimacroutils.SetForegroundWindow(hndle)
            if not result:
                print('switch to %s failed, delete name: %s'% (hndle, countOrApp))
                del self.namedtaskDict[countOrApp]
                self.setList('namedtask', list(self.namedtaskDict.keys()))
                return None
            return result
        if countBack:
            self.lastTaskCount = countBack
            if self.doTasksWithWindowsKey:
                self.goto_task_winkey(countBack)
            else:
                action('TASK %s'% countBack)
            result = unimacroutils.getProgInfo()
            #print 'after action task %s, time: %.2f'% (countBack, (time.time()-t))
            return result
        if countOrApp in appList:
            value = self.ini.getList('application', countOrApp)
            if len(value) == 1:
                app = value[0]
                self.app = app
                result = action("BRINGUP %s"% app)
                return result
##            print 'after bringup: %s'% app
            if len(value) == 2:
                #application is known, but click by number!!
                appName = value[0]
                countBack = value[1]
                self.lastTaskCount = countBack
                if self.doTasksWithWindowsKey:
                    self.goto_task_winkey(countBack)
                else:
                    action('TASK %s'% countBack)
                for _ in range(30):
                    # 40 x 0.1: 4 seconds...
                    prog_info = unimacroutils.getProgInfo()
                    if prog_info.prog == appName:
                        break
                    prog_info.classname = unimacroutils.getClassName()
                    if prog_info.classname == "TaskListThumbnailWnd":
                        return 1  # more items already available
                    unimacroutils.Wait()
                else:
                    print('application not detected in foreground: %s'% appName)
                    return None
        else:
            print('_tasks, no valid entry for gotoTask: %s'% countOrApp)
            return None
        result = unimacroutils.getProgInfo()
        return None

    def gotoDocument(self, count):
        """go to the specified document, by number or application name, return proginfo, or None if task was not found
        
        """
        result = action('DOCUMENT %s'% count)
        print('result of gotoDocument %s: %s'% (count, result))
        return result
    
    def goto_task_winkey(self, number):
        """switch to task with number, via the windows key"""
    ##    print 'action: goto task: %s'% number
        prog_info = unimacroutils.getProgInfo()
        prog, title = prog_info.prog, prog_info.title
        if prog == 'explorer' and not title:
            keystroke('{esc}')
            unimacroutils.shortWait()
        try:
            count = int(number)
        except ValueError:
            print('goto_task_winkey, invalid number: %s'% number)
            return None
        if not count:
            print('goto_task_winkey, invalid number: %s'% number)
            return None
        if count == 10:
            count=0
        elif count > 10:
            print('goto_task_winkey, pass on to "TASK %s", number > 9'% count)
            return action('TASK %s'% count)
        
        self.doWinKey('b')
        actions.do_VW()
        self.doWinKey(str(number))
        print('self.winkeyDown: %s'% self.winkeyDown)
        return None
        
    # def rule_taskposition(self, words):
    #     """
    #     #commands for positioning (moving) and resizing tasks: 
    #     (<taskswitch>|task) position <directionplus> 
    #     """
    #     # position task in one of the directions, width/height
    #     # did also optional percent with <directionplus> <percent> switched off for the moment
    #     # optional percentage of work area
    #     self.taskmoveresize = 'position'
    #     print('----task position')
        
    # def rule_taskmove(self, words):
    #     """(<taskswitch> | task) move <directionplus>;
    #     """
    #     
    #     # removed all the angle, pixels, centimeters, inches, percent out, too complicated for daily use I think...
    #     # optional a specification
    #     self.taskmoveresize = 'move'
    #     print('----task move')
    
    #def rule_taskresize(self, words):
    #    """(<taskswitch> | (task)) (stretch|shrink)
    #            (<directionplus>|<angle>)
    #               [<pixels>|<centimeters>|<millimeters>|<inches>|<percent>]
    #    """
    #    # resize a task in direction or angle
    #    # with optional specification
    #    if self.hasCommon(words, 'stretch'):
    #        self.taskmoveresize = 'stretch'
    #    if self.hasCommon(words, 'shrink'):
    #        self.taskmoveresize = 'shrink'
    #    print '----task resize: %s'% self.taskmoveresize
    #    # action in gotResult
        
    def rule_taskresizeinfo(self, words):
        """give resize info
        """
        winHndle = win32gui.GetForegroundWindow()
        canBeResized = monitorfunctions.window_can_be_resized(winHndle)
        print('resize info %s'% canBeResized)
            
    
    # rules inside position, move, stretch, shrink:
    # too precise, can be switched on again later...
    #def subrule_angle(self, words):
    #    """ 
    #    #directional specifications:
    #    {degrees} degrees
    #    """
    #    #get the move or resize angle
    #    deg = self.getNumberFromSpoken(words[0])
    #    self.directionplus = deg%360
    #    print 'angle/direction: %s'% self.directionplus
        
    # def subrule_directionplus(self, words):
    #     '{directionplus}'
    #     #get the direction
    #     direction = words[0]
    #     if direction in self.directionsplus:
    #         self.directionplus = self.directionsplus[direction]
    #     else:
    #         self.directionplus = direction
    #     print('direction: (plus) %s'% self.directionplus)

    def rule_monitorfocus(self, words):
        'monitor {monitors}'
        # put focus on monitor
        # target is the index of the MONITOR_HNDLES list of monitorfunctions.py.
        # for 3 monitors 0, 1 or 2. Which is which has to be tried, hopefully the
        # list is consistent
        monitorIndex = int(self.getFromInifile(words[1], 'monitors'))
        # print 'ask for monitor: %s'% monitorIndex WORK AREA
        rect = monitorfunctions.get_monitor_rect_work(monitorIndex)
        # print 'rect: %s'% repr(rect)
        if not rect:
            print('rule_monitorfocus, no position rectangle found')
            return
        mx, my = unimacroutils.relToCoord(0.5, rect[0], rect[2]), unimacroutils.relToCoord(0.01, rect[1], rect[3])
        unimacroutils.doMouse(0, 0, mx, my, mouse='left')
        unimacroutils.visibleWait()
        # relative and relative to current monitor work area:
        unimacroutils.doMouse(1, 4, 0.5, 0.5, mouse="noclick")
        unimacroutils.visibleWait()
        # mx, my = unimacroutils.relToCoord(0.5, rect[0], rect[2]), unimacroutils.relToCoord(0.5, rect[1], rect[3])
        # unimacroutils.doMouse(0, 0, mx, my, mouse='noclick')
        # actions.doAction("RMP(3, 0.3, 0.3, mouse='noclick')")

    #def subrule_pixels(self, words):
    #    """
    #    # size specifications:
    #    {pixelcount} | {pixelcount} pixels
    #    """
    #    #get the amount in pixels (optional word)
    #    amountList = self.getNumberFromSpoken(words)
    #    if len(amountList) == 1:
    #        self.amount = amountList[0]
    #    else:
    #        raise ValueError("task grammar, rule pixels, more amounts found: %s"% amountList)
    #    self.units = 'pixels'
    #    print 'pixels: %s'% self.amount
    #
    #def subrule_centimeters(self, words):
    #    '{sizecount} centimeters'
    #    """get the amount in centimeters"""
    #    nCm = self.getNumberFromSpoken(words[0])
    #    self.units = 'pixels'
    #    self.amount = nCm * self.dotsperinch * 1.0 / 2.54 + 0.5
    #    print 'pixels(%s cm): %s'% (nCm, self.amount)
    #
    #def subrule_millimeters(self, words):
    #    '{sizecount} millimeters'
    #    #get the amount in millimeters
    #    nMm = self.getNumberFromSpoken(words[0])
    #    self.units = 'pixels'
    #    self.amount = nMm * self.dotsperinch * 1.0 / 25.4 + 0.5
    #    print 'pixels(%s mm): %s'% (nMm, self.amount)
    #    
    #def subruleprog_infonches(self, words):
    #    '{sizecount} inches'
    #    #get the amount in inches
    #    nInches = self.getNumberFromSpoken(words[0])
    #    self.units = 'pixels'
    #    self.amount = nInches * self.dotsperinch
    #    print 'pixels(%s mm): %s'% (nInches, self.amount)
        
    def rule_gettaskposition(self, words):
        """
        #commands for recording taskbar positions:
        get task position {taskcount}
        """
        # getting the task positions (use with 1 and with another number)
        # position mouse on task number or clock and speak the command
        # june 2016: removed the obsolete clock commands.
        # first time only, or after changes of taskbar position
        count = self.getNumberFromSpoken(words[-1])
        x, y = natlink.getCursorPos()
        if count in self.taskCounts:
            print('%s, setting task position: %s'% (self.name, count))
        else:
            print('%s, not a valid "count" for setting task position: %s'% (self.name, count))
            return
        if count == 1:
            print('setting mouseX1: ', x)
            print('setting mouseY1: ', y)
            setPosition('mousex1', x)
            setPosition('mousey1', y)
        else:
            mousex1 = getPosition('mousex1')
            mousey1 = getPosition('mousey1')
            
            mouseXdiff = int((x - mousex1 )/(count-1))
            mouseYdiff = int((y - mousey1 )/(count-1))
            print('setting mouseXdiff: ', mouseXdiff)
            print('setting mouseYdiff: ', mouseYdiff)
            setPosition('mousexdiff', mouseXdiff)
            setPosition('mouseydiff', mouseYdiff)
        
    def rule_getdocumentposition(self, words):
        """
        #commands for recording a document position, application specific
        get document position {documentcount}
        """
        # getting the task positions (use with 1 and with another number)
        # position mouse on task number or clock and speak the command
        # first time only, or after changes of taskbar position
        prog_info = unimacroutils.getProgInfo()
        prog, title = prog_info.prog, prog_info.title
        if not prog:
            print('%s, no valid program for setting document position: %s (title:%s)'% (self.name, prog, title))
            return
        count = self.getNumberFromSpoken(words[-1])
        x, y = natlink.getCursorPos()
        if count in self.documentCounts:
            print('%s, setting document position %s for program: %s'% (self.name, count, prog))
        else:
            print('%s, cannot set document position "%s" for program: %s (invalid count)'% (self.name, count, prog))
            return
        if count == 1:
            print('setting mouseX1: ', x)
            print('setting mouseY1: ', y)
            setPosition('mousex1', x, prog=prog)
            setPosition('mousey1', y, prog=prog)
        else:
            mousex1 = getPosition('mousex1', prog=prog)
            mousey1 = getPosition('mousey1', prog=prog)
            
            mouseXdiff = int((x - mousex1 )/(count-1))
            mouseYdiff = int((y - mousey1 )/(count-1))
            print('setting mouseXdiff: ', mouseXdiff)
            print('setting mouseYdiff: ', mouseYdiff)
            setPosition('mousexdiff', mouseXdiff, prog=prog)
            setPosition('mouseydiff', mouseYdiff, prog=prog)


    #
    def gotResults_fileswitch(self, words, fullResults):
        """switch file to the named application
        mark cursor
        close in current app
        open in new app
        goto top
        goto cursor
        
        if filename cannot be got the process is stopped

        """
        fileName = actions.getPathOfOpenFile()
        if fileName:
            print('fileName: %s'% fileName)
        else:
            self.DisplayMessage('cannot switch, filename cannot be established')
            return
            
        while fileName.endswith("*"):
            fileName = fileName[:-1]
        actions.putCursor()
        action("<<filesave>>")
        action("W")
        print("saved, now close: %s"% fileName)
        action("<<documentclose>>")
        newApp = words[-1]
        newApp = self.ini.get('application', newApp, newApp)
           
        if not action("BRINGUP %s"% newApp):
            self.DisplayMessage('cannot switch, application %s cannot be brought to front'% newApp)
            return

        # fileopen in emacs relies on maximised state of emacs, cannot be sure here:
        if newApp in ('emacs', 'voicecode'):
            action("{ctrl+x}{ctrl+f}; {shift+home}{del}")
        else:
            action("<<fileopen>>")
        action("W")
        keystroke(os.path.normpath(fileName))
        action("VW")
        keystroke("{enter}")
        action("<<topdocument>>")
        actions.findCursor()
        action("<<filesave>>")

    def gotResults(self, words, fullResults):
        """actions for task move, resize, position done here
        
        relevant data collected in other gotResult_... functions
        
        """
        if self.giveName:
            if self.namedictated:
                self.namedtaskDict[self.namedictated] = self.giveName
                self.setList('namedtask', list(self.namedtaskDict.keys()))
                print('name "%s" can now be used for switching to a task'% self.namedictated)
            else:
                print('giveName, no self.namedictated given...')
                
        if self.hadStartMenu:
            self.doStartMenu()
            if self.letters:
                keystroke(self.letters)

        if self.taskmoveresize:
            print('taskmoveresize: %s'% self.taskmoveresize)
            winHndle = win32gui.GetForegroundWindow()
            #print 'amount: %s'% self.amount
            #print 'units: %s'% self.units
            if self.directionplus in self.directions:
                direction = self.directions[self.directionplus]
            else:
                direction = self.directionplus
            print('direction spoken: %s, direction taken: %s'% (self.directionplus, direction))
            canBeResized = monitorfunctions.window_can_be_resized(winHndle)
            if self.taskmoveresize == 'position':
                self.doTaskPositionCommand(winHndle, canBeResized, direction, self.amount, self.units)
            else:
                # move, stretch or shrink:
                if self.amount:
                    amount = self.amount
                    units = self.units or 'pixels'
                elif self.taskmoveresize == 'move':
                    amount = 1
                    units = 'relative'
                elif self.taskmoveresize == 'stretch':
                    amount = 1
                    units = 'relative'
                elif self.taskmoveresize == 'shrink':
                    amount = 0.5
                    units = 'relative'
                else:
                    print('_tasks, gotResults, invalid taskmoveresize: %s'% self.taskmoveresize)
                    return

                print('amount (adjusted): %s'% amount)
                print('units: (adjusted) %s'% units)
    
                try:
                    func = getattr(monitorfunctions, f'{self.taskmoveresize}_window', None)
                    if func:
                        print('doing func: {func}')
                        func(winHndle, direction=self.directionplus, amount=amount, units=units)
                    else:
                        print('invalid value for taskmoveresize: %s (could not find function)'% self.taskmoveresize)                
                except ValueError:
                    print('error in monitorfunctions.%s_window'% self.taskmoveresize)
                    print(sys.exc_info()[1])
#keepinside=None, keepinsideall=1, monitor=None):

    def addTaskNameToDict(self, moduleInfo):
        """add name of task to the namedTasDict is requirements are met
        
        no return, change is made in namedTaskDict and the setList is done as well
        
        1. must be pythonw (already tested in gotBegin)
        2. title must hold " --- " (applications of QH)
        3. check for abbrevs list in spokenforms, to make the command pronouncable
        
        """
        _prog, title, hndle = moduleInfo
        if title.find(" --- ") == -1:
            return
        name = title.split()[0]
        name = name.strip("~ ")

        if not name:
            return
        gotChanges = None
        
        name = self.spokenforms.abbrev2spoken.get(name, name)
        if name and isinstance(name, list):
            for n in name:
                gotChanges = self.checkNamedTaskDict(n, hndle) or gotChanges
        else:
            gotChanges = self.checkNamedTaskDict(name, hndle)
        if gotChanges:
            print('set namedtask list: %s'% list(self.namedtaskDict.keys()))
            self.setList('namedtask', list(self.namedtaskDict.keys()))
            
    
    def checkNamedTaskDict(self, name, hndle):
        """insert name, hndle in self.namedTaskDict if not yet there
        
        return 1 if change has been made.    
    
        """
        if self.namedtaskDict.get(name, None) == hndle:
            return None  # name already in dict and same hndle
        self.namedtaskDict[name] = hndle
        return 1

    def doWinKey(self, keys, keepdown=None):
        """do a winkey and return the window info of the resulting window
        """
        winkey = win32con.VK_LWIN         # 91
        keyup = win32con.KEYEVENTF_KEYUP  # 2
        win32api.keybd_event(winkey, 0, 0, 0)  # key down
        classInfo = None
        try:
            actions.do_SSK(keys)
            actions.do_VW()
            classInfo = unimacroutils.getClassName()
        finally:
            if classInfo == 'TaskListThumbnailWnd':
                #print 'keep logo key down'
                self.winkeyDown = 1
            else:
                win32api.keybd_event(winkey, 0, keyup, 0)  # key up
        
        
    
    # taken from _shownumbersplus, go to a subtask in stacked taskbar
    # number from 0, (in _shownumbersplus first window is click 0)
    def doAlternativeClick(self, className, num):
        """instead of click perform another (keystrokes) action
        
        if self.doTasksWithWindowsKey, do this with repeated winkey clicks
        """
        taskNum = self.lastTaskCount
        if type(num) in (bytes, str):
            num = int(num)
        if className == "TaskListThumbnailWnd":
            #print 'tasks, doAlternative click: %s'% num
            if self.doTasksWithWindowsKey:
                if not (taskNum and self.winkeyDown):
                    print('cannot complete windows command: taskNum: %s, winkeyDown: %s'% (taskNum, self.winkeyDown))
                    self.cancelMode()
                    return
                if num > 1:
                    keystr = str(taskNum)
                    for _ in range(num-1):
                        actions.do_SSK(keystr)
                elif num < 0:
                    num2 = -num
                    keystr = '{shift+%s}'% taskNum
                    for _i in range(num2):
                        actions.do_SSK(keystr)
                self.cancelMode()
                return

            # mode with mouse on taskbar position:
            taskbarPosition = monitorfunctions.get_taskbar_position() # notices changes on the fly!
            # which way to scroll through a stack of taskbar buttons:
            initialKeys = dict(left='{down}', top='{right}',
                               right='{left}{down}', bottom='{right}')
            scrollKeys = dict(left='down', top='right',
                               right='down', bottom='right')
            initialKeysReverse = dict(right='{left}', bottom='{left}',
                               left='{right}', top='{left}')
            scrollKeysReverse = dict(right='up', bottom='left',
                               left='up', top='left')
            try:
                if num > 0:
                    initial = initialKeys[taskbarPosition]
                    scrollKey = scrollKeys[taskbarPosition]
                elif num < 0:
                    initial = initialKeysReverse[taskbarPosition]
                    scrollKey = scrollKeysReverse[taskbarPosition]
                    num = -num
                else:
                    raise ValueError("_tasks, doAlternativeClick, number may NOT be 0")
            except KeyError:
                print('key not found "taskbarPosition": %s'% taskbarPosition)
                return
            # go to first of popup windows (depending on taskbar location)
            #print 'taskbarPosition: %s'% taskbarPosition
            #print 'initial: %s, scroll: %s'% (initial[1:], scrollKey)
            action(initial)
            if num > 1:
                num -= 1
                action('{%s %s}'% (scrollKey, num))
            action('{enter}')
        else:
            print('should not call doAlternativeClick for className: %s'% className)


    def doTaskPositionCommand(self, winHndle, canBeResized, direction, amount, units):
        """do the task position command with enclosed parameters
        """
        if amount and units == 'percent':
            amount = amount / 100.0
        elif amount:
            print('amount: %s'% amount)
        else:
            amountx = self.splitLeftRight
            amounty = self.splitTopDown
        xwidth = ywidth = 1.0
        if canBeResized:
            print('setting default pos to 0.0')
            xpos = ypos = 0.0
        else:
            print('no resize, setting default pos to 0.5')
            xpos = ypos = 0.5
        if direction.find('left') >= 0:
            xpos = 'left'
            xwidth = amountx
        elif direction.find('right') >= 0:
            xpos = 'right'
            xwidth = 1.0 - amountx
        elif direction.startswith('center'):
            xpos = 'center'
            if amountx >= 0.5:
                xwidth = amountx
            else:
                xwidth = 1.0 - amountx

        if direction.find('up') >= 0:
            ypos = 'up'
            ywidth = amounty
        elif direction.find('down') >= 0:
            ypos = 'down'
            ywidth = 1.0 - amounty
        elif direction.endswith('center'):
            ypos = 'center'
            if amounty >= 0.5:
                ywidth = amounty
            else:
                ywidth = 1.0 - amounty
    
        if not canBeResized:
            xwidth = ywidth = None
        print('restore_window:pos: %s, %s, width: %s, %s'% \
                (xpos, ypos, xwidth, ywidth))
        func = monitorfunctions.restore_window
        func(winHndle, xpos=xpos, ypos=ypos,
                xwidth=xwidth, ywidth=ywidth)
    #
    #
    #def convertString(self, t, toCode):
    #      """convert to unix style (\n) or dos/windows style \n\r sucks..."""
    #      CR = chr(10)
    #      LF = chr(13)
    #      CRLF = CR + LF
    #      CRCRLF = CR + CR + LF
    #      if toCode == 'unix':
    #          return t.replace(CRLF, LF)
    #      elif toCode in ('windows', 'dos'):
    #          t = t.replace(LF, CRLF)
    #          t = t.replace(CRCRLF, CRLF)
    #          t = t.replace(CRLF+CRLF, CRLF)
    #          return t 
        
    def gotResults_removecursor(self, words, fullResults):
        """goto marked cursor, in case is was left in file switch actions"""
        actions.findCursor()

    def doTaskAction(self, actionWord, mouseOnPosition=0):
        """do action on 

     if mouseOnPosition, instead of {alt+space} a right click is done
        """
        if actionWord:
            act = self.ini.get('taskaction', actionWord)
            if act:
                #unimacroutils.visibleWait()
                action(act)
            else:
                print('no action for taskaction: %s'% actionWord)
        

    def fillInstanceVariables(self):
        """fills the necessary instance variables

        positions should be in actions.ini, and can be got by the "task position" commands
        
        """
        self.maxTaskNumber = self.ini.getInt('general', 'max task number') or 20
        self.maxWindowNumber = self.ini.getInt('general', 'max window number') or 9
        self.maxDocumentNumber = self.ini.getInt('general', 'max document number') or 20
        self.centerMouse = self.ini.getBool('general', 'center mouse')
        #self.dotsperinch = self.ini.getInt('general', 'screen dots per inch') or 90
        
        self.doTasksWithWindowsKey = self.ini.getBool('general', 'do taskswitch with windows key')
        if self.doTasksWithWindowsKey:
            print('do taskswitch with windows key enabled')
        # positioning not in middle (for Arnoud):
        self.splitLeftRight = self.ini.getFloat('general', 'split left right') or 0.5
        self.splitTopDown = self.ini.getFloat('general', 'split top down') or 0.5
        
        # search commands (for Arnoud):
        self.enableSearchCommands = self.ini.getBool('general', 'enable search commands') 
        if self.enableSearchCommands:
            print('_task, enable search commands')

        # special windows from pythonw (QH)
        self.makeTaskNamesPythonw = self.ini.getBool('general', 'make task names pythonw') 
        if self.makeTaskNamesPythonw:
            print('_task, make task names pythonw (option for QH)')
        
        allApps = self.ini.get('application')

        # contstruct 
        switchApps = self.ini.getList('general', 'switchapps')
        self.switchApps = []
        if switchApps:
            # construct items for switchApps list, must be either
            # the keyword or the value of one of the items of
            # allApps (the list of applications in the inifile
            
            allAppsDict = {k: self.ini.get('application', k) for k in allApps}

            for sApp in switchApps:
                if sApp in allApps:
                    self.switchApps.append(sApp)
                else:
                    foundSpoken = 0
                    for k,v in list(allAppsDict.items()):
                        if v == sApp:
                            foundSpoken = 1
                            self.switchApps.append(k)
                    if not foundSpoken:
                        print('application not valid for switchapps: "%s", no written or spoken form in section [applications])'% sApp)

    def getSelectedText(self):
        """gets a copy of the selection, otherwise ""
        """
        unimacroutils.saveClipboard()
        action("<<copy>>")
        unimacroutils.Wait()
        t = unimacroutils.getClipboard()
        unimacroutils.restoreClipboard()
        return t.strip()

    def cancelMode(self):
        # ending exclusive mode if in it, also ending the mode it is in
        winkey = win32con.VK_LWIN         # 91
        keyup = win32con.KEYEVENTF_KEYUP  # 2
        if self.winkeyDown:
            print('tasks, cancelMode, release WINKEY')
            win32api.keybd_event(winkey, 0, keyup, 0)  # key up
            self.winkeyDown = 0

            

def unload():
    #pylint:disable = W0603, E0601
    global thisGrammar
    if thisGrammar:
        thisGrammar.unload()
    thisGrammar = None

def changeCallback(Type,Args):
    # not active without special version of natlinkmain:
    if ((Type == 'mic') and (Args=='on')):
        return   # check WAS in natlinkmain...
    if thisGrammar:
        thisGrammar.cancelMode()

# standard stuff Joel
if __name__.find('.') == -1:
    # this is caught when this module is imported by the loader (when Dragon/Natlink starts)
    thisGrammar = ThisGrammar()
    thisGrammar.initialize()

# testing via test_tasks.py



"""Unimacro __init__


Note there will be a global variable created in the unimacro module 'ulogger' which is Logging.Logger object named 'natlink.unimacro'
You can always access it by name.  It is created in _control.py.

"""
import os
import sys

def folders_logger_name() -> str:
    return "natlink.unimacro.folders"

def control_logger_name() -> str : 
    return "natlink.unimacro.control"

def logname() -> str:
    """ Returns the name of the unimacro logger."""
    return "natlink.unimacro"

__version__ = '4.1.6'



#
# Python Macro Language for Dragon NaturallySpeaking
#   (c) Copyright 1999 by Joel Gould
#   Portions (c) Copyright 1999 by Dragon Systems, Inc.
#
# unittestIniGrammar.py
#   This script performs some basic tests of the NatLink 
#   with the IniGrammar subclass.
#   This also tests features of IniGrammar, BrowsableGrammar and GrammarX
#   being the chain of grammars leading up to GrammarBase
#
#   especially test translate feature of the IniGrammar.
#   note special tests which should be made active with care, when in debugging mode for catching specific commands!
#   see def xxxx.....
#
import sys
import unittest
import types
import os
import os.path
import time
import traceback        # for printing exceptions
import TestCaseWithHelpers
import natlink
from natlinkcore import loader
from natlinkcore import natlinkstatus
from dtactions.unimacro import inivars
from natlinkcore.gramparser import GrammarError, GrammarSyntaxError

status = natlinkstatus.NatlinkStatus()

from natlinkcore import natlinkutils
from dtactions.unimacro import unimacroutils
from dtactions.unimacro import unimacroutils
import unimacro.natlinkutilsbj as natbj
from dtactions.unimacro.unimacroactions import doAction as action
from dtactions.unimacro.unimacroactions import doAction as action

class TestError(Exception):
    pass
ExitQuietly = 'ExitQuietly'

def getBaseFolder(globalsDict=None):
    """get the folder of the calling module.

    either sys.argv[0] (when run direct) or
    __file__, which can be empty. In that case take the working directory
    """
    globalsDictHere = globalsDict or globals()
    baseFolder = ""
    if globalsDictHere['__name__']  == "__main__":
        baseFolder = os.path.split(sys.argv[0])[0]
        print('baseFolder from argv: %s'% baseFolder)
    elif globalsDictHere['__file__']:
        baseFolder = os.path.split(globalsDictHere['__file__'])[0]
        print('baseFolder from __file__: %s'% baseFolder)
    if not baseFolder or baseFolder == '.':
        baseFolder = os.getcwd()
        print('baseFolder was empty, take wd: %s'% baseFolder)
    return baseFolder

thisDir = getBaseFolder(globals())

natconnectOption = 1 # or 1 for threading, 0 for not. Seems to make difference
                     # with spurious error (if set to 1), missing gotBegin and all that...
logFileName = os.path.join(thisDir, "testresult.txt")

#---------------------------------------------------------------------------
# these testruns should be runs when natlink is NOT enabled, but Unimacro is. You establish this by configuring natlink etc.
# as you wish, and then Disable NatLink.
# This debugger run will start_natlink so has connection with all the Unimacro grammars. 
# Restart Dragon and kill extra Dragon instances in the tasks application before rerunning the whole procedure.
# Debugging done this way by Quintijn, September 2013
class UnittestIniGrammar(TestCaseWithHelpers.TestCaseWithHelpers):
    connected = 0
    def setUp(self):
        if not natlink.isNatSpeakRunning():
            raise TestError('NatSpeak is not currently running')
        if not self.connected:
            self.connect()
            self.user = natlink.getCurrentUser()[0]
            self.setMicState = "off"
            self.connected = 1

    def tearDown(self):
        # do nothing, leaving natlink on!
        return
        #try:
        #    # give message:
        #    self.setMicState = "off"
        #    # kill things
        #    self.clearTestFiles().2000
        #finally:
        #    self.disconnect()

    def makeInifile(self, inifilename, iniSpec=None):
        """make in subdirectory inifiles_test a new inifile of this name
        
        """
        inifilepath = os.path.join(thisDir, 'test_inifiles', inifilename)
        ini = inivars.IniVars(inifilepath)
        if iniSpec:
            ini.delete()
            ini.fromDict(iniSpec)
            ini.write()
        return ini
        
    def connect(self):
        # start with 1 for thread safety when run from pythonwin:
        print('connecting')
        natlink.natConnect(natconnectOption)
        # next line should be enabled when testing the testNumbersGrammar functions in debug mode:
        # otherwise disable!
        print('connected natlink (Unimacro)')
        pass

    #def disconnect(self):
        #"""is not called in this test file!"""
        #natlink.natDisconnect()
        
    def log(self, t):
        # only log to file:
        log(t)

    def wait(self, t=1):
        time.sleep(t)


    #---------------------------------------------------------------------------
    # This utility subroutine executes a Python command and makes sure that
    # an exception (of the expected type) is raised.  Otherwise a TestError
    # exception is raised

    def doTestForException(self, exceptionType,command,localVars={}):
        try:
            exec(command,globals(),localVars)
        except exceptionType:
            return
        raise TestError('Expecting an exception to be raised calling '+command)

    def doTestActiveRules(self, gram, expected):
        """gram must be a grammar instance, sort the rules to be expected and got
        """
        expected.sort()
        got = gram.activeRules
        got.sort()
        self.assertEqual(expected, got,
                         'Active rules not as expected:\nexpected: %s, got: %s'%
                         (expected, got))

    def doTestFuncReturn(self, expected,command,localVars=None):
        # account for different values in case of [None, 0] (wordFuncs)
        if localVars == None:
            actual = eval(command)
        else:
            actual = eval(command, globals(), localVars)

        if actual != expected:
            time.sleep(1)
        self.assertEqual(expected, actual, 'Function call "%s" returned unexpected result\nExpected: %s, got: %s'%
                          (command, expected, actual))
    def doTestNumberRecognition(self, words,numberResult):
        try:
            natlink.recognitionMimic(words)
        except natlink.MimicFailed:
            print('failed to recognise %s as testnumber'% words)
            return
        return 1 # success

    def doTestRecognition(self, words,shouldWork=1, log=None):
        if shouldWork:
            natlink.recognitionMimic(words)
            if log:
                self.log("recognised: %s"% words)
        else:
            self.doTestForException(natlink.MimicFailed,"natlink.recognitionMimic(words)",locals())
            if log:
                self.log("did not recognise (as expected): %s"% words)

    
    #def tttestBrowseAllGrammarsControl(self):
    #    """go into the browse all grammars via control grammar
    #       found errors in the BrowseGrammar procedures by debugging this function!!
    #    """
    #    loadedFiles = natlinkmain.loadedFiles
    #    import _control
    #    phrase = ['show', 'all', 'grammars']  # enx
    #    phrase = ['toon', 'alle', 'grammaticaas']  # nld dutch
    #    natlink.recognitionMimic(phrase)
    #    pass
    
    def xxxtestNumbersSimpleGrammarNld(self):
        """test the numbers grammar rules as written in natlinkutilsbj
        
        attach to the _number grammar, with language dutch (translated word for number is "getal").
        (this grammar must be in the Unimacro directory)
        please remove (if applicable) _number extended.py
        
        handle with care, run or debug ONLY this function, with NatLink disabled, see next function (below)
        get the seqsAndRules and fullResults by uncommenting 2 print lines in natlinkutils.py, lin 670.
        
        """
        natlinkmain.start_natlink()
        print('started natlink (Unimacro)')
        _number = __import__("_number simple")
        numbersgrammar = _number.thisGrammar
        
        ## 302000303:
        words = ['getal', 'drie', 'honderd', 'twee', 'miljoen', 'drie', 'honderd', 'drie']
        seqsAndRules = [(['getal'], 'testnumber'), (['drie'], '__1to99'), (['honderd'], '__hundreds'),
                        (['twee'], '__1to99'), (['miljoen'], '__millions'),
                        (['drie'], '__1to99'), (['honderd'], '__hundreds'), (['drie'], '__1to99')]
        fullResults = [('getal', 'testnumber'), ('drie', '__1to99'), ('honderd', '__hundreds'),
                        ('twee', '__1to99'), ('miljoen', '__millions'), ('drie', '__1to99'),
                        ('honderd', '__hundreds'), ('drie', '__1to99')]
        
        # now call the different functions in the resultsCallback procedure:
        numbersgrammar.callIfExists( 'gotResultsInit', (words, fullResults) )
        numbersgrammar.callRuleResultsFunctions(seqsAndRules, fullResults)
        numbersgrammar.callIfExists( 'gotResults', (words, fullResults) )
        numberExpected = '302000303'
        numbersgrammar.collectNumber()
        numberGot = numbersgrammar.number
        self.assertEqual(numberExpected, numberGot,
                          "numbers grammar gives unexpected result\nexpected: %s\ngot: %s"% (numberExpected, numberGot))
        #

        # next:
        # 3200005: (foutje: 3201005)
        words =  ['getal', 'drie', 'miljoen', 'twee', 'honderd', 'duizend', 'vijf']

        seqsAndRules = [(['getal'], 'testnumber'), (['drie'], '__1to99'), (['miljoen'], '__millions'),
                        (['twee'], '__1to99'), (['honderd'], '__hundreds'),
                        (['duizend'], '__thousands'), (['vijf'], '__1to99')]

        fullResults = [('getal', 'testnumber'), ('drie', '__1to99'), ('miljoen', '__millions'),
                       ('twee', '__1to99'), ('honderd', '__hundreds'),
                       ('duizend', '__thousands'), ('vijf', '__1to99')]

        numbersgrammar.callIfExists( 'gotResultsInit', (words, fullResults) )
        numbersgrammar.callRuleResultsFunctions(seqsAndRules, fullResults)
        numbersgrammar.callIfExists( 'gotResults', (words, fullResults) )
        numberExpected = '3200005'
        numbersgrammar.collectNumber()
        numberGot = numbersgrammar.number
        self.assertEqual(numberExpected, numberGot,
                          "numbers grammar gives unexpected result\nexpected: %s\ngot: %s"% (numberExpected, numberGot))
      
        # 324700024    (foutje: 3201005)
        words =  ['getal', 'drie', 'honderd', 'vierentwintig', 'miljoen', 'zeven', 'honderd', 'duizend', 'vierentwintig']

        seqsAndRules = [(['getal'], 'testnumber'), (['drie'], '__1to99'), (['honderd'], '__hundreds'), (['vierentwintig'], '__1to99'),
                        (['miljoen'], '__millions'),
                        (['zeven'], '__1to99'), (['honderd'], '__hundreds'), (['duizend'], '__thousands'), (['vierentwintig'], '__1to99')]

        fullResults = [('getal', 'testnumber'), ('drie', '__1to99'), ('honderd', '__hundreds'), ('vierentwintig', '__1to99'),
                       ('miljoen', '__millions'),
                       ('zeven', '__1to99'), ('honderd', '__hundreds'), ('duizend', '__thousands'), ('vierentwintig', '__1to99')]

        numbersgrammar.callIfExists( 'gotResultsInit', (words, fullResults) )
        numbersgrammar.callRuleResultsFunctions(seqsAndRules, fullResults)
        numbersgrammar.callIfExists( 'gotResults', (words, fullResults) )
        numberExpected = '324700024'
        numbersgrammar.collectNumber()
        numberGot = numbersgrammar.number
        self.assertEqual(numberExpected, numberGot,
                          "numbers grammar gives unexpected result\nexpected: %s\ngot: %s"% (numberExpected, numberGot))
       #
       
        # 3,5
        words =  ['getal', 'drie', 'komma', 'vijf']
        seqsAndRules = [(['getal'], 'testnumber'), (['drie'], '__0to99'), (['komma'], 'float'), (['vijf'], '__0to99')]
        fullResults = [('getal', 'testnumber'), ('drie', '__0to99'), ('komma', 'float'), ('vijf', '__0to99')]

        numbersgrammar.callIfExists( 'gotResultsInit', (words, fullResults) )
        numbersgrammar.callRuleResultsFunctions(seqsAndRules, fullResults)
        numbersgrammar.callIfExists( 'gotResults', (words, fullResults) )
        numberExpected = '3,5'
        numbersgrammar.collectNumber()
        numberGot = numbersgrammar.number
        self.assertEqual(numberExpected, numberGot,
                          "numbers grammar gives unexpected result\nexpected: %s\ngot: %s"% (numberExpected, numberGot))
       
    def ttttestNumbersExtendedFixedLengthNld(self):
        """test the numbers grammar rules as written in natlinkutilsbj
        
        attach to the _number grammar, with language dutch (translated word for number is "getal").
        (this grammar must be in the Unimacro directory)
        please remove (if applicable) _number_simple.py
        
        handle with care, run or debug ONLY this function, with NatLink disabled, see next function (below)
        
        
        """
        natlinkmain.start_natlink()
        print('started natlink (Unimacro)')
        _number = __import__("_number extended")
        numbersgrammar = _number.thisGrammar
        
        ## example fixed length punt (point) + four digits,
        ## to be dictated as "two thousand" ("twee duizend")
        words = ['punt', 'twee', 'duizend']
        fullResults = [('punt', 'numberaztwo'), ('twee', '__2to9before'), ('duizend', '__thousandslimited')]
        seqsAndRules= [(['punt'], 'numberaztwo'), (['twee'], '__2to9before'), (['duizend'], '__thousandslimited')]

        
        
        # ## 8623, note we test this inside the rule numberazone, initially developed for a user in Amsterdam Zuid
        # words = ['zesentachtig', 'drie', 'drie\xebntwintig']  #
        # seqsAndRules = [(['zesentachtig'], 'numberazone'), (['drie\xebntwintig'], 'numbertwodigits')]
        # fullResults = [('zesentachtig', 'numberazone'), ('drie\xebntwintig', 'numbertwodigits')]
        
        # now call the different functions in the resultsCallback procedure:
        numbersgrammar.callIfExists( 'gotResultsInit', (words, fullResults) )
        numbersgrammar.callRuleResultsFunctions(seqsAndRules, fullResults)
        numbersgrammar.callIfExists( 'gotResults', (words, fullResults) )
        numberExpected = '.2000'
        numbersgrammar.collectNumber()
        numberGot = numbersgrammar.number
        self.assertEqual(numberExpected, numberGot,
                          "numbers grammar gives unexpected result\nexpected: %s\ngot: %s"% (numberExpected, numberGot))
        #

        # next:
        # 3200005: (foutje: 3201005)
        words =  ['getal', 'drie', 'miljoen', 'twee', 'honderd', 'duizend', 'vijf']

        seqsAndRules = [(['getal'], 'testnumber'), (['drie'], '__1to99'), (['miljoen'], '__millions'),
                        (['twee'], '__1to99'), (['honderd'], '__hundreds'),
                        (['duizend'], '__thousands'), (['vijf'], '__1to99')]

        fullResults = [('getal', 'testnumber'), ('drie', '__1to99'), ('miljoen', '__millions'),
                       ('twee', '__1to99'), ('honderd', '__hundreds'),
                       ('duizend', '__thousands'), ('vijf', '__1to99')]

        numbersgrammar.callIfExists( 'gotResultsInit', (words, fullResults) )
        numbersgrammar.callRuleResultsFunctions(seqsAndRules, fullResults)
        numbersgrammar.callIfExists( 'gotResults', (words, fullResults) )
        numberExpected = '3200005'
        numbersgrammar.collectNumber()
        numberGot = numbersgrammar.number
        self.assertEqual(numberExpected, numberGot,
                          "numbers grammar gives unexpected result\nexpected: %s\ngot: %s"% (numberExpected, numberGot))
      
        # 324700024    (foutje: 3201005)
        words =  ['getal', 'drie', 'honderd', 'vierentwintig', 'miljoen', 'zeven', 'honderd', 'duizend', 'vierentwintig']

        seqsAndRules = [(['getal'], 'testnumber'), (['drie'], '__1to99'), (['honderd'], '__hundreds'), (['vierentwintig'], '__1to99'),
                        (['miljoen'], '__millions'),
                        (['zeven'], '__1to99'), (['honderd'], '__hundreds'), (['duizend'], '__thousands'), (['vierentwintig'], '__1to99')]

        fullResults = [('getal', 'testnumber'), ('drie', '__1to99'), ('honderd', '__hundreds'), ('vierentwintig', '__1to99'),
                       ('miljoen', '__millions'),
                       ('zeven', '__1to99'), ('honderd', '__hundreds'), ('duizend', '__thousands'), ('vierentwintig', '__1to99')]

        numbersgrammar.callIfExists( 'gotResultsInit', (words, fullResults) )
        numbersgrammar.callRuleResultsFunctions(seqsAndRules, fullResults)
        numbersgrammar.callIfExists( 'gotResults', (words, fullResults) )
        numberExpected = '324700024'
        numbersgrammar.collectNumber()
        numberGot = numbersgrammar.number
        self.assertEqual(numberExpected, numberGot,
                          "numbers grammar gives unexpected result\nexpected: %s\ngot: %s"% (numberExpected, numberGot))
       #
       
        # 3,5
        words =  ['getal', 'drie', 'komma', 'vijf']
        seqsAndRules = [(['getal'], 'testnumber'), (['drie'], '__0to99'), (['komma'], 'float'), (['vijf'], '__0to99')]
        fullResults = [('getal', 'testnumber'), ('drie', '__0to99'), ('komma', 'float'), ('vijf', '__0to99')]

        numbersgrammar.callIfExists( 'gotResultsInit', (words, fullResults) )
        numbersgrammar.callRuleResultsFunctions(seqsAndRules, fullResults)
        numbersgrammar.callIfExists( 'gotResults', (words, fullResults) )
        numberExpected = '3,5'
        numbersgrammar.collectNumber()
        numberGot = numbersgrammar.number
        self.assertEqual(numberExpected, numberGot,
                          "numbers grammar gives unexpected result\nexpected: %s\ngot: %s"% (numberExpected, numberGot))
       


       
    def xxxxtestNumbersSimpleGrammarEnx(self):
        """test the numbers grammar rules as written in natlinkutilsbj
        
        
        attach to the _number simple.py grammar, with language english (this grammar must be in the Unimacro directory)
        please remove (if applicable) _number extended.py

        !!Enable ONLY this test function, disable NatLink (leaving Unimacro and possibly Vocola enabled)
        !!enable the "start_natlink" line in def connect and
        then restart Dragon and run this module.
        
        Note the test function writes the number into this file, so leave the cursor here or after a comment
        before running the function!! Or you comment out the line self.outputNumber() (line 99) in _number simple.py.

        """
        natlinkmain.start_natlink()
        print('started natlink (Unimacro)')
        
        _number = __import__("_number simple")
        numbersgrammar = _number.thisGrammar
        
        
        ## testing leading 0: 03, oh three:
        words = ['Number', 'oh', 'three']
        seqsAndRules = [(['Number'], 'testnumber'), (['oh', 'three'], '__0to99')]
        fullResults =[('Number', 'testnumber'), ('oh', '__0to99'), ('three', '__0to99')]

        # now call the different functions in the resultsCallback procedure:
        numbersgrammar.callIfExists( 'gotResultsInit', (words, fullResults) )
        numbersgrammar.callRuleResultsFunctions(seqsAndRules, fullResults)
        numbersgrammar.callIfExists( 'gotResults', (words, fullResults) )
        numberExpected = '03'
        numberGot = numbersgrammar.number
        self.assertEqual(numberExpected, numberGot,
                          "numbers grammar gives unexpected result\nexpected: %s\ngot: %s"% (numberExpected, numberGot))


        ## 302000303:
        words = ['Number', 'three', 'hundred', 'two', 'million', 'three', 'hundred', 'three']
        seqsAndRules = [(['Number'], 'testnumber'), (['three'], '__1to99'), (['hundred'], '__hundreds'), (['two'], '__1to99'),
                        (['million'], '__millions'),
                        (['three'], '__1to99'), (['hundred'], '__hundreds'), (['three'], '__1to99')]
        fullResults =[('Number', 'testnumber'), ('three', '__1to99'), ('hundred', '__hundreds'), ('two', '__1to99'),
            ('million', '__millions'),
            ('three', '__1to99'), ('hundred', '__hundreds'), ('three', '__1to99')]

        # now call the different functions in the resultsCallback procedure:
        numbersgrammar.callIfExists( 'gotResultsInit', (words, fullResults) )
        numbersgrammar.callRuleResultsFunctions(seqsAndRules, fullResults)
        numbersgrammar.callIfExists( 'gotResults', (words, fullResults) )
        numberExpected = '302000303'
        numberGot = numbersgrammar.number
        self.assertEqual(numberExpected, numberGot,
                          "numbers grammar gives unexpected result\nexpected: %s\ngot: %s"% (numberExpected, numberGot))


    def xxxtestCalculatorNld(self):
        """test the numbers grammar rules in calculator grammar
        
        Note the precautions stated above around line 390.
        This is for testing the _calculator grammar in the debugger

        """
        natlinkmain.start_natlink()
        print('started natlink (Unimacro)')
        
        _calc = __import__("_calculator")
        calcgrammar = _calc.thisGrammar
        
        
        ## testing  simple calculation (dutch):
        seqsAndRules = [(['bereken'], 'calculation'), (['drie'], '__0to9'), (['plus'], 'operator'),
            (['vijf'], '__0to9'), (['is gelijk aan'], 'calculation')]
        fullResults = [('bereken', 'calculation'), ('drie', '__0to9'), ('plus', 'operator'), ('vijf', '__0to9'), ('is gelijk aan', 'calculation')]
        words = ['bereken', 'drie', 'plus', 'vijf', 'is gelijk aan']
        words = ['reken uit', 'drie', 'keer', 'min', 'vier', 'is gelijk aan']
        seqsAndRules = [(['reken uit'], 'calcnormal'), (['drie'], '__0to9'), (['keer', 'min'], 'operator'), (['vier'], '__0to9'), (['is gelijk aan'], 'calcnormal')]
        fullResults = [('reken uit', 'calcnormal'), ('drie', '__0to9'), ('keer', 'operator'), ('min', 'operator'), ('vier', '__0to9'), ('is gelijk aan', 'calcnormal')]

        # now call the different functions in the resultsCallback procedure
        # 
        calcgrammar.callIfExists( 'gotResultsInit', (words, fullResults) )
        calcgrammar.callRuleResultsFunctions(seqsAndRules, fullResults)
        calcgrammar.callIfExists( 'gotResults', (words, fullResults) )
        numberExpected = '03'
        numberGot = calcgrammar.number
        self.assertEqual(numberExpected, numberGot,
                          "calc grammar gives unexpected result\nexpected: %s\ngot: %s"% (numberExpected, numberGot))

    def xxxxtestKeystrokesGrammarEnx(self):
        """test the keystrokes grammar

        with language english (this grammar must be in the Unimacro directory)
        the inactive mode seemed not to be working. Deactivate was not enough. Hard to test.
        But hopefully solved, Quintijn, September 26, 2013...

        !!Enable ONLY this test function, disable NatLink (leaving Unimacro and possibly Vocola enabled)
        !!enable the "start_natlink" line in def connect and
        then restart Dragon and run this module.
        
        Note the test function writes the number into this file, so leave the cursor here or after a comment
        before running the function!! Or you comment out the line self.outputNumber() (line 99) in _number simple.py.

        """
        natlinkmain.start_natlink()
        print('started natlink (Unimacro)')
        import _keystrokes
        kgrammar = _keystrokes.thisGrammar
        
        
        # the Backspace gives an error:        
        words = ['Backspace']
        fullResults = [('Backspace', 'repkey')]
        seqsAndRules = [(['Backspace'], 'repkey')]

        # now call the different functions in the resultsCallback procedure:
        moduleInfo = natlink.getCurrentModule()
        # initialise things in gotBegin
        kgrammar.callIfExists( 'gotBegin', (moduleInfo,) )
        # and run the rules:
        kgrammar.callIfExists( 'gotResultsInit', (words, fullResults) )
        kgrammar.callRuleResultsFunctions(seqsAndRules, fullResults)
        kgrammar.callIfExists( 'gotResults', (words, fullResults) )
        pass
        ##

    #def xxxxtestFolderGrammarSiteSpecificationNld(self):
    #    """test the import of some site module. Specific option for Quintijn
    #    
    #    !!Enable ONLY this test function, disable NatLink (leaving Unimacro and possibly Vocola enabled)
    #    !!enable the "start_natlink" line in def connect and
    #    then restart Dragon and run this module.
    #    
    #    """
    #    natlinkmain.start_natlink()
    #    print 'started natlink (Unimacro)'
    #    import _folders
    #    fgrammar = _folders.thisGrammar
    #    fgrammar.getSiteInstance('sg')
    # if this error happens again, ensure utilsqh and inivars are same version as in
    # QH's miscqh folder. (Sept 2013)
    #    pass


    def xxxxtestControlGrammarNld(self):
        """test the reload of the control grammar
        
        The obsolete words seem to be reversed.

        !!Enable ONLY this test function, disable NatLink (leaving Unimacro and possibly Vocola enabled)
        !!enable the "start_natlink" line in def connect and
        then restart Dragon and run this module.
        
        Note the test function writes the number into this file, so leave the cursor here or after a comment
        before running the function!! Or you comment out the line self.outputNumber() (line 99) in _number simple.py.Zwitserland kiest Rooks hallo

        This test was because at reload time the self.gramSpec instance variable was used for translation instead of
        self.__class__.gramSpec.
        
        Next lines of test are not very sensible. Problem was at natlinkutilsbj line 2324, QH sept 2013

        """
        natlinkmain.start_natlink()
        print('started natlink (Unimacro)')
        import _control
        cgrammar = _control.utilGrammar  # why this name?
        print('gramSpec: %s'% cgrammar.gramSpec)
        words = ['schakel', 'in', 'toetsen']
        fullResults = [('schakel', 'switch'), ('in', 'switch'), ('toetsen', 'switch')]
        seqsAndRules = [(['schakel', 'in', 'toetsen'], 'switch')]

        moduleInfo = natlink.getCurrentModule()
        cgrammar.checkInifile()

        # now call the different functions in the resultsCallback procedure:
        print('after second initialise gramSpec: %s'% cgrammar.gramSpec)
        pass
        ##

    
    def tttestGrammarRepeatedTranslationsAndLoads(self):
        """test a simple grammar with rules that are translated, do multiple loads to check stability
        
        """
        class TestGrammar(natbj.IniGrammar):
            
            gramSpec = '<reload> exported = reload translated grammar;'
            
            def __init__(self):
                natbj.IniGrammar.__init__(self)

        testGram = TestGrammar()
        iniSpec = {'grammar name': {'name':'reload test'},
                   'grammar words':
            {'reload': ['opnieuw laden'],
 'rule': ['regel'],
 'test': ['test', 'testing'],
 'translated': ['vertaald', 'translated']}}
        testGram.ini = self.makeInifile('testreloadtranslated.ini', iniSpec=iniSpec)

        # test the translation words dictionary from fake inifile (iniSpec) above:        
        expGramSpec = ['<four> exported = rule four;']
        expGramScanList = []
        expDict =  {'reload': ['opnieuw laden'],
 'rule': ['regel'],
 'test': ['test', 'testing'],
 'translated': ['vertaald', 'translated']}
        gotDict = testGram.getDictOfGrammarWordsTranslations()
        self.assert_equal(expDict, gotDict,  "translated words dict not as expected")
        
        # do the translation step:
        gotGramSpec = testGram.translateGrammar(testGram.gramSpec)
        expGramSpec =  "<reload> exported = 'opnieuw laden' (vertaald|translated) grammar;"
        self.assert_equal(expGramSpec, gotGramSpec,  "translated grammar was not as expected")
        expGramWords =  {'reload': ['opnieuw laden'], 'translated': ['vertaald', 'translated']}
        gotGramWords = testGram.gramWords
        self.assert_equal(expGramWords, gotGramWords,  "gramwords dict after translation not as expected")

        gotGotGramSpec = testGram.translateGrammar(testGram.gramSpec)
        self.assert_equal(gotGramSpec, gotGotGramSpec, "twice translated grammar should be the same as once translated")

        gotGotGramWords = testGram.gramWords
        self.assert_equal(expGramWords, gotGotGramWords,  "gramwords dict after double translation not as expected")
                   
        # clean up
        testGram.unload()


    def tttestGrammarTranslation(self):
        """test a simple grammar with rules that are translated
        
        either other language or with synonyms
        """
        class TestGrammar(natbj.IniGrammar):
            
            gramSpec = '<four> exported = rule four;'
            
            def __init__(self):
                natbj.IniGrammar.__init__(self)

        testGram = TestGrammar()
        iniSpec = {'grammar name': {'name':'test rule four'},
                   'grammar words':
            {'four':'four;for', 'rule': 'regel'}}
        testGram.ini = self.makeInifile('testrulefour.ini', iniSpec=iniSpec)

        # test the translation words dictionary from fake inifile (iniSpec) above:        
        expGramSpec = ['<four> exported = rule four;']
        expGramScanList = []
        expDict =  {'four': ['four', 'for'], 'rule': ['regel']}
        gotDict = testGram.getDictOfGrammarWordsTranslations()
        self.assert_equal(expDict, gotDict,  "translated words dict not as expected")
        
        # do the translation step:
        gotGramSpec = testGram.translateGrammar(testGram.gramSpec)
        expGramSpec = '<four> exported = regel (four|for);';
        self.assert_equal(expGramSpec, gotGramSpec,  "translated grammar was not as expected")
        
        ### now try small changes in the rule ( no extra parens needed):
        testGram.gramSpec = '<fourparens> exported = rule (four);'
        gotGramSpec = testGram.translateGrammar(testGram.gramSpec)
        expGramSpec = '<fourparens> exported = regel (four|for);';
        self.assert_equal(expGramSpec, gotGramSpec,  "translated no extra parens needed grammar was not as expected")
        ## try other variants of no parens needed
        testGram.gramSpec = '<fourparens2> exported = rule (four|five|six);'
        gotGramSpec = testGram.translateGrammar(testGram.gramSpec)
        expGramSpec = '<fourparens2> exported = regel (four|for|five|six);';
        self.assert_equal(expGramSpec, gotGramSpec,  "translated no extra parens needed grammar was not as expected")
        # another:
        testGram.gramSpec = '<fourparens3> exported = rule [three|four|five|six];'
        gotGramSpec = testGram.translateGrammar(testGram.gramSpec)
        expGramSpec = '<fourparens3> exported = regel [three|four|for|five|six];';
        self.assert_equal(expGramSpec, gotGramSpec,  "translated no extra parens needed grammar was not as expected")
        # another:
        testGram.gramSpec = '<fourparens4> exported = rule [four];'
        gotGramSpec = testGram.translateGrammar(testGram.gramSpec)
        expGramSpec = '<fourparens4> exported = regel [four|for];';
        self.assert_equal(expGramSpec, gotGramSpec,  "translated no extra parens needed grammar was not as expected")
        
        
        expGramWords =  {'four': ['four', 'for'], 'rule': ['regel']}
        gotGramWords = testGram.gramWords
        self.assert_equal(expGramWords, gotGramWords,  "gramwords dict after translation not as expected")
        
               
        # clean up
        testGram.unload()


    def tttestGrammarTranslationComplicated(self):
        """test a little bit more involved grammar with rules that are translated
        
        either other language or with synonyms
        """
        class TestGrammar(natbj.IniGrammar):
            
            gramSpec = '''#demo translation
<one> exported = rule <two> ['a lot'] EXTRA;
#more rules:
<two> = Exported with {two} {one} [and ['a lot'|little]] more; # additional comment
#try to do my best'''
            
            def __init__(self):
                natbj.IniGrammar.__init__(self)

        testGram = TestGrammar()
        iniSpec = {'grammar name': {'name':'test rule one two'},
                   'grammar words':
            {'four':'four;for',
             'rule': 'regel',
             'one': 'een',
             'Exported': 'Uitgevoerd|"een beetje Afgevoerd"',
             'extra': 'Extra',
             'A lot':'veel',
             'more': 'more',
             'little': '"een beetje"|weinig'}}
        testGram.ini = self.makeInifile('testruleonetwo.ini', iniSpec=iniSpec)

        # test the translation words dictionary from fake inifile (iniSpec) above:        
        expDict =     {'a lot': ['veel'],
 'exported': ['Uitgevoerd', 'een beetje Afgevoerd'],
 'extra': ['Extra'],
 'four': ['four', 'for'],
 'little': ['een beetje', 'weinig'],
 'one': ['een'],
 'rule': ['regel']}
        gotDict = testGram.getDictOfGrammarWordsTranslations()
        self.assert_equal(expDict, gotDict,  "translated words dict not as expected")
        
        # do the translation step:
        gotGramSpec = testGram.translateGrammar(testGram.gramSpec)
        expGramSpec = '''#demo translation\n<one> exported = regel <two> ['veel'] Extra;\n#more rules:\n<two> = (Uitgevoerd|'een beetje Afgevoerd') with {two} {one} [and ['veel'|'een beetje'|weinig]] more; # additional comment\n#try to do my best'''
        self.assert_equal(expGramSpec, gotGramSpec,  "translated grammar was not as expected")
        testGram.load(gotGramSpec)
        expRules = ['one']
        gotRules = testGram.validRules
        self.assert_equal(expRules, gotRules, "translated grammar does not load into the correct rules")
               
        # clean up
        testGram.unload()

    def testGrammarTranslationInifileHandling(self):
        """test the changes of translated words and obsolete words in the inifile
        
        """
        class TestGrammar(natbj.IniGrammar):
            
            gramSpec = '''#grammar with simple translation
<simple> exported = rule one;
'''
            
            def __init__(self):
                natbj.IniGrammar.__init__(self)

        testGram = TestGrammar()
        iniSpec = {'grammar name': {'name':'test inifile handling'},
                   }
        testGram.ini = self.makeInifile('testinifilehandling.ini', iniSpec=iniSpec)

        # test the translation words dictionary from fake inifile (iniSpec) above:        
        expDict =  {}
        gotDict = testGram.getDictOfGrammarWordsTranslations()
        self.assert_equal(expDict, gotDict,  "translated words dict not as expected")
        
        # no translation yet:
        gotGramSpec = testGram.translateGrammar(testGram.gramSpec)
        expGramSpec = None  # no translation yet
        self.assert_equal(expGramSpec, gotGramSpec,  "no translated grammar at this moment")
        
        expIniSpec =    {'grammar name': {'name': 'test inifile handling'}}

        gotIniSpec = testGram.ini.toDict()
        self.assert_equal(expIniSpec, gotIniSpec, "non-translated grammar does not produce expected ini file")
        
        ## make an obsolete word and translate again:
        testGram.gramSpec = '''<simple> exported = rule;'''
        gotGramSpec = testGram.translateGrammar(testGram.gramSpec)
        expGramSpec = None  # no translation yet

        self.assert_equal(expGramSpec, gotGramSpec,  "no translated grammar at this moment")
      
        expIniSpec =    {'grammar name': {'name': 'test inifile handling'}}
        gotIniSpec = testGram.ini.toDict()
        self.assert_equal(expIniSpec, gotIniSpec, "non-translated grammar does not produce expected ini file")


        ## now introduce a synonym
        testGram.gramSpec = '''<simple> exported = rule one;'''
        iniSpec = {'grammar name': {'name':'test inifile handling'},
                'grammar words': {'one': 'synonym'}
                   }
        testGram.ini = self.makeInifile('testinifilehandling.ini', iniSpec=iniSpec)
        
        gotGramSpec = testGram.translateGrammar(testGram.gramSpec)
        expGramSpec =  '<simple> exported = rule synonym;'

        self.assert_equal(expGramSpec, gotGramSpec,  "no translated grammar at this moment")
      
      
        ## note the language is here "zyx", so not enx of nld:      
        expIniSpec =        {'grammar name': {'name': 'test inifile handling'},
 'grammar non translated words': {'info1': 'These grammar words can be translated.',
                                   'info2': 'See http://qh.antenna.nl/unimacro/features/translations for more info',
                                   'words': ['rule']},
 'grammar words': {'one': 'synonym'}}

        gotIniSpec = testGram.ini.toDict()
        self.assert_equal(expIniSpec, gotIniSpec, "translated grammar does not produce expected ini file")


        ## remove the translated word:
        testGram.gramSpec = '''<simple> exported = rule;'''
        # go on with the previous inifile        
        
        gotGramSpec = testGram.translateGrammar(testGram.gramSpec)
        expGramSpec =  None # again no translation

        self.assert_equal(expGramSpec, gotGramSpec,  "no translated grammar at this moment")
      
        expIniSpec =       {'grammar name': {'name': 'test inifile handling'},
 'grammar non translated words': {'info1': 'These grammar words can be translated.',
                                   'info2': 'See http://qh.antenna.nl/unimacro/features/translations for more info',
                                   'words': ['rule']},
 'grammar obsolete words': {'one': ['synonym']}}

        gotIniSpec = testGram.ini.toDict()
        self.assert_equal(expIniSpec, gotIniSpec, "non translated grammar with obsolete words does not produce expected ini file")

        ## make translated word active again, catching the translation:
        testGram.gramSpec = '''<simple> exported = rule one;'''
        # go on with the previous inifile        
        
        gotGramSpec = testGram.translateGrammar(testGram.gramSpec)
        # if no translation, None is returned
        expGramSpec =  None  ### u'<simple> exported = rule synonym;'

        self.assert_equal(expGramSpec, gotGramSpec,  "no translated grammar at this moment")
      
        expIniSpec =    {'grammar name': {'name': 'test inifile handling'},
 'grammar non translated words': {'info1': 'These grammar words can be translated.',
                                   'info2': 'See http://qh.antenna.nl/unimacro/features/translations for more info',
                                   'words': ['rule']},
 'grammar obsolete words': {'one': ['synonym']}}
        gotIniSpec = testGram.ini.toDict()
        self.assert_equal(expIniSpec, gotIniSpec, "non translated grammar with obsolete words does not produce expected ini file")


        ## make translated words active again, catching the translation:
        testGram.gramSpec = '''<simple> exported = rule one;'''
        ##
        ## augment ini spec with items that can be removed:
        ## identical words are NOT remembered!!
        ##
        iniSpec =     {'grammar name': {'name': 'test inifile handling'},
'grammar non translated words': {'info1': 'These grammar words can be changed if you want synonyms for them.',
                                  'info2': 'See http://qh.antenna.nl/unimacro/features/translations for more info',
                                  'words': ['rule']},
'grammar words': {'one': ['synonym', 'one'], 'rule': ['rule']},
'grammar obsolete words': {'two': ['keep'], 'three': ['three|also keep'], 'four': ['four']}}

        testGram.ini = self.makeInifile('testinifilehandling.ini', iniSpec=iniSpec)
        
        gotGramSpec = testGram.translateGrammar(testGram.gramSpec)
        expGramSpec =  '<simple> exported = rule (synonym|one);'

        self.assert_equal(expGramSpec, gotGramSpec,  "translated grammar with synonyms not correct")
      
        expIniSpec =     {'grammar name': {'name': 'test inifile handling'},
 'grammar non translated words': {'info1': 'These grammar words can be translated.',
                                   'info2': 'See http://qh.antenna.nl/unimacro/features/translations for more info',
                                   'words': ['rule']},
 'grammar obsolete words': {'three': ['three|also keep'], 'two': ['keep']},
 'grammar words': {'one': ['synonym', 'one']}}

        gotIniSpec = testGram.ini.toDict()
        self.assert_equal(expIniSpec, gotIniSpec, "non translated grammar with obsolete words does not produce expected ini file")


        ## now translated also the last word of the grammar:
        testGram.ini.set('grammar words', 'rule', 'regel')
        testGram.ini.write()
        pass
        gotGramSpec = testGram.translateGrammar(testGram.gramSpec)
        expGramSpec =  '<simple> exported = rule (synonym|one);'
        
        expIniSpec =  {'grammar name': {'name': 'test inifile handling'},
 'grammar obsolete words': {'three': ['three|also keep'], 'two': ['keep']},
 'grammar words': {'one': ['synonym', 'one'], 'rule': 'regel'}}

        gotIniSpec = testGram.ini.toDict()
        self.assert_equal(expIniSpec, gotIniSpec, "non translated grammar with obsolete words does not produce expected ini file")
 

    def tttestGrammarTranslationTwoLinesDoubleWord(self):
        """test a little bit more involved grammar, try if additional newlines are caugth and double words correctly translated
        
        """
        class TestGrammar(natbj.IniGrammar):
            
            gramSpec = '''#test comment with translation
<doubleword> exported = rule <two> ['right click'|"double click"];
<two> = 'another rule' exported;
'''
            
            def __init__(self):
                natbj.IniGrammar.__init__(self)

        testGram = TestGrammar()
        iniSpec = {'grammar name': {'name':'test rule doubleword'},
                   'grammar words':
            {'right click':'rechts klik; rechter klik',
             'double click': 'dubbelklik',
             'rule': 'regel',
             'another rule': 'nog een regel'}}
        testGram.ini = self.makeInifile('testruledoubleword.ini', iniSpec=iniSpec)

        # test the translation words dictionary from fake inifile (iniSpec) above:        
        expDict =        {'another rule': ['nog een regel'],
 'double click': ['dubbelklik'],
 'right click': ['rechts klik', 'rechter klik'],
 'rule': ['regel']}
        gotDict = testGram.getDictOfGrammarWordsTranslations()
        self.assert_equal(expDict, gotDict,  "translated words dict not as expected")
        
        # do the translation step:
        gotGramSpec = testGram.translateGrammar(testGram.gramSpec)
        expGramSpec =  '''#test comment with translation
<doubleword> exported = regel <two> ['rechts klik'|'rechter klik'|"dubbelklik"];
<two> = 'nog een regel' exported;'''

        self.assert_equal(expGramSpec, gotGramSpec,  "translated grammar was not as expected")
        testGram.load(gotGramSpec)
        expRules = ['doubleword']
        gotRules = testGram.validRules
        self.assert_equal(expRules, gotRules, "translated grammar does not load into the correct rules")
        
        # use for testing the showInifile function:
        #testGram.showInifile()
        
               
        # clean up
        testGram.unload()


def log(t):
    """log to print and file if present

    note print depends on the state of natlink: where is goes or disappears...
    I have no complete insight is this, but checking the logfile afterwards
    always works (QH)
    """
    print(t)
    if logFile:
        logFile.write(t + '\n')
    
#---------------------------------------------------------------------------
# run
#
# This is the main entry point.  It will connect to NatSpeak and perform
# a series of tests.  In the case of an error, it will cleanly disconnect
# from NatSpeak and print the exception information,
def dumpResult(testResult, logFile):
    """dump into 
    """
    if testResult.wasSuccessful():
        mes = "all succesful"
        logFile.write(mes)
        return
    logFile.write('\n--------------- errors -----------------\n')
    for case, tb in testResult.errors:
        logFile.write('\n---------- %s --------\n'% case)
        logFile.write(tb)
        
    logFile.write('\n--------------- failures -----------------\n')
    for case, tb in testResult.failures:
        logFile.write('\n---------- %s --------\n'% case)
        logFile.write(tb)

    


logFile = None

def run():
    global logFile, natconnectOption
    logFile = open(logFileName, "w")
    log("log messages to file: %s"% logFileName)
    log('starting unittestNatlink')
    # trick: if you only want one or two tests to perform, change
    # the test names to her example def tttest....
    # and change the word 'test' into 'tttest'...
    # do not forget to change back and do all the tests when you are done.
    suite = unittest.makeSuite(UnittestIniGrammar, 'test')
##    natconnectOption = 0 # no threading has most chances to pass...
    log('\nstarting tests with threading: %s\n'% natconnectOption)
    result = unittest.TextTestRunner().run(suite)
    dumpResult(result, logFile=logFile)
    
    logFile.close()

if __name__ == "__main__":
    run()
